# stdlib
from datetime import datetime
from typing import TYPE_CHECKING
from typing import cast

# relative
from ...serde.serializable import serializable
from ...store.linked_obj import LinkedObject
from ..context import AuthedServiceContext

if TYPE_CHECKING:
    # relative
    from .notifications import Notification


class EmailTemplate:
    @staticmethod
    def email_title(notification: "Notification", context: AuthedServiceContext) -> str:
        return ""

    @staticmethod
    def email_body(notification: "Notification", context: AuthedServiceContext) -> str:
        return ""


@serializable(canonical_name="PasswordResetTemplate", version=1)
class PasswordResetTemplate(EmailTemplate):
    @staticmethod
    def email_title(notification: "Notification", context: AuthedServiceContext) -> str:
        return "Password Reset Requested"

    @staticmethod
    def email_body(notification: "Notification", context: AuthedServiceContext) -> str:
        user_service = context.server.get_service("userservice")

        user = user_service.get_by_verify_key(notification.to_user_verify_key)
        if not user:
            raise Exception("User not found!")

        user.reset_token = user_service.generate_new_password_reset_token(
            context.server.settings.pwd_token_config
        )
        user.reset_token_date = datetime.now()

        result = user_service.stash.update(
            credentials=context.credentials, user=user, has_permission=True
        )
        if result.is_err():
            raise Exception("Couldn't update the user password")

        head = """<head>
            <style>
                body {
                    font-family: Arial, sans-serif;
                    background-color: #f4f4f4;
                    color: #333;
                    margin: 0;
                    padding: 0;
                }
                .container {
                    max-width: 600px;
                    margin: 50px auto;
                    background-color: #ffffff;
                    padding: 20px;
                    border-radius: 8px;
                    box-shadow: 0 2px 4px rgba(0, 0, 0, 0.1);
                }
                h1 {
                    color: #333;
                    text-align: center;
                }
                p {
                    font-size: 16px;
                    line-height: 1.5;
                }
                .button {
                    display: block;
                    width: 200px;
                    margin: 30px auto;
                    padding: 10px;
                    background-color: #007BFF;
                    color: #fff;
                    text-align: center;
                    text-decoration: none;
                    border-radius: 4px;
                }
                .footer {
                    text-align: center;
                    font-size: 12px;
                    color: #aaa;
                    margin-top: 20px;
                }
            </style>
        </head>"""
        body = f"""<body>
            <div class="container">
                <h1>Password Reset</h1>
                <p>We received a request to reset your password. Your new temporary token is:</p>
                <h1>{user.reset_token}</h1>
                <p> Use
                    <code style="color: #FF8C00;background-color: #f0f0f0;font-size: 12px;">
                        syft_client.reset_password(token='{user.reset_token}', new_password=*****)
                    </code>.
                to reset your password.</p>
                <p>If you didn't request a password reset, please ignore this email.</p>
            </div>
        </body>"""
        return f"""<html>{head} {body}</html>"""


@serializable(canonical_name="OnboardEmailTemplate", version=1)
class OnBoardEmailTemplate(EmailTemplate):
    @staticmethod
    def email_title(notification: "Notification", context: AuthedServiceContext) -> str:
        return f"Welcome to {context.server.name} server!"

    @staticmethod
    def email_body(notification: "Notification", context: AuthedServiceContext) -> str:
        user_service = context.server.get_service("userservice")
        admin_name = user_service.get_by_verify_key(
            user_service.admin_verify_key()
        ).name

        head = (
            f"""
        <head>
            <title>Welcome to {context.server.name}</title>
        """
            + """
            <style>
                body {
                    font-family: Arial, sans-serif;
                    background-color: #f4f4f4;
                    color: #333;
                    line-height: 1.6;
                }
                .container {
                    max-width: 600px;
                    margin: 20px auto;
                    padding: 20px;
                    background: #fff;
                }
                h1 {
                    color: #0056b3;
                }
                .feature {
                    background-color: #e7f1ff;
                    padding: 10px;
                    margin: 10px 0;
                    border-radius: 5px;
                }
                .footer {
                    text-align: center;
                    font-size: 14px;
                    color: #aaa;
                }
            </style>
        </head>
        """
        )

        body = f"""
        <body>
            <div class="container">
                <h1>Welcome to {context.server.name} server!</h1>
                <p>Hello,</p>
                <p>We're thrilled to have you on board and
                excited to help you get started with our powerful features:</p>

                <div class="feature">
                    <h3>Remote Data Science</h3>
                    <p>Access and analyze data from anywhere, using our comprehensive suite of data science tools.</p>
                </div>

                <div class="feature">
                    <h3>Remote Code Execution</h3>
                    <p>Execute code remotely on private data, ensuring flexibility and efficiency in your research.</p>
                </div>

                <!-- Add more features here if needed -->

                <p>Explore these features and much more within your account.
                If you have any questions or need assistance, don't hesitate to reach out.</p>

                <p>Cheers,</p>
                <p>{admin_name}</p>

                <div class="footer">
                    This is an automated message, please do not reply directly to this email. <br>
                    For assistance, please contact our support team.
                </div>
            </div>
        </body>
        """
        return f"""<html>{head} {body}</html>"""


@serializable(canonical_name="RequestEmailTemplate", version=1)
class RequestEmailTemplate(EmailTemplate):
    @staticmethod
    def email_title(notification: "Notification", context: AuthedServiceContext) -> str:
        notification.linked_obj = cast(LinkedObject, notification.linked_obj)
        request_obj = notification.linked_obj.resolve_with_context(context=context).ok()

        return f"Datasite {context.server.name}: A New Request ({str(request_obj.id)[:4]}) has been received!"

    @staticmethod
    def email_body(notification: "Notification", context: AuthedServiceContext) -> str:
        notification.linked_obj = cast(LinkedObject, notification.linked_obj)
        request_obj = notification.linked_obj.resolve_with_context(context=context).ok()

        head = """
        <head>
            <title>Access Request Notification</title>
            <style>
                body {
                    font-family: Arial, sans-serif;
                    background-color: #f4f4f4;
                    color: #333;
                    padding: 20px;
                }
                .container {
                    max-width: 600px;
                    margin: 0 auto;
                    background: #fff;
                    padding: 20px;
                    border-radius: 8px;
                    box-shadow: 0 0 10px rgba(0, 0, 0, 0.1);
                }
                .header {
                    font-size: 24px;
                    color: #333;
                    text-align: center;
                }
                .content {
                    font-size: 16px;
                    line-height: 1.6;
                }

                .request-card {
                    background-color: #ffffff;
                    border: 1px solid #ddd;
                    padding: 15px;
                    margin-top: 20px;
                    border-radius: 8px;
                    box-shadow: 0 2px 4px rgba(0,0,0,0.1);
                }

                .request-header {
                    font-size: 18px;
                    color: #333;
                    margin-bottom: 10px;
                    font-weight: bold;
                }

                .request-content {
                    font-size: 14px;
                    line-height: 1.5;
                    color: #555;
                }
                .badge {
                    padding: 4px 10px;
                    border-radius: 15px;
                    color: white;
                    font-weight: bold;
                    margin: 10px;
                    box-shadow: 0px 4px 6px rgba(0,0,0,0.1);
                    text-transform: uppercase;
                }

                .yellow {
                    background-color: #fdd835;
                }

                .green {
                    background-color: #6dbf67;
                }

                .red {
                    background-color: #f7786b;
                }

                .button {
                    display: block;
                    width: max-content;
                    background-color: #007bff;
                    color: white;
                    padding: 10px 20px;
                    text-align: center;
                    text-color: white;
                    border-radius: 5px;
                    text-decoration: none;
                    font-weight: bold;
                    margin: 20px auto;
                }
                .footer {
                    text-align: center;
                    font-size: 14px;
                    color: #aaa;
                }
            </style>
        </head>"""

        body = f"""
        <body>
            <div class="container">
                <div class="header">
                    Request Notification
                </div>
                <div class="content">
                    <p>Hello,</p>
                    <p>A new request has been submitted and requires your attention.
                    Please review the details below:</p>

                    <div class="request-card">
                        <div class="request-header">Request Details</div>
                        <div class="request-content">

                            <p><strong>ID:</strong> {request_obj.id}</p>
                            <p>
                            <strong>Submitted By:</strong>
                            {request_obj.requesting_user_name} {request_obj.requesting_user_email or ""}
                            </p>
                            <p><strong>Date:</strong> {request_obj.request_time}</p>
                        <div style="display: flex"><p><strong>Status:</strong><div class="badge yellow">{
                            request_obj.status.name
                        }</div></div>
                        <p><strong>Changes:</strong>{
                            ",".join([change.__class__.__name__ for change in request_obj.changes])
                        }</p>
                        </div>
                    </div>
                    <p>If you did not expect this request or have concerns about it,
                    please contact our support team immediately.</p>
                </div>
                <div class="footer">
                    This is an automated message, please do not reply directly to this email. <br>
                    For assistance, please contact our support team.
                </div>
            </div>
        </body>
        """
        return f"""<html>{head} {body}</html>"""


@serializable(canonical_name="RequestUpdateEmailTemplate", version=1)
class RequestUpdateEmailTemplate(EmailTemplate):
    @staticmethod
    def email_title(notification: "Notification", context: AuthedServiceContext) -> str:
        return f"Datasite {context.server.name}: {notification.subject}"

    @staticmethod
    def email_body(notification: "Notification", context: AuthedServiceContext) -> str:
        notification.linked_obj = cast(LinkedObject, notification.linked_obj)
        request_obj = notification.linked_obj.resolve_with_context(context=context).ok()
        badge_color = "red" if request_obj.status.name == "REJECTED" else "green"
        head = """
        <head>
            <title>Access Request Notification</title>
            <style>
                body {
                    font-family: Arial, sans-serif;
                    background-color: #f4f4f4;
                    color: #333;
                    padding: 20px;
                }
                .container {
                    max-width: 600px;
                    margin: 0 auto;
                    background: #fff;
                    padding: 20px;
                    border-radius: 8px;
                    box-shadow: 0 0 10px rgba(0, 0, 0, 0.1);
                }
                .header {
                    font-size: 24px;
                    color: #333;
                    text-align: center;
                }
                .content {
                    font-size: 16px;
                    line-height: 1.6;
                }

                .request-card {
                    background-color: #ffffff;
                    border: 1px solid #ddd;
                    padding: 15px;
                    margin-top: 20px;
                    border-radius: 8px;
                    box-shadow: 0 2px 4px rgba(0,0,0,0.1);
                }

                .request-header {
                    font-size: 18px;
                    color: #333;
                    margin-bottom: 10px;
                    font-weight: bold;
                }

                .badge {
                    padding: 4px 10px;
                    border-radius: 15px;
                    color: white;
                    font-weight: bold;
                    margin: 10px;
                    box-shadow: 0px 4px 6px rgba(0,0,0,0.1);
                    text-transform: uppercase;
                }

                .yellow {
                    background-color: #fdd835;
                }

                .green {
                    background-color: #6dbf67;
                }

                .red {
                    background-color: #f7786b;
                }

                .request-content {
                    font-size: 14px;
                    line-height: 1.5;
                    color: #555;
                }

                .button {
                    display: block;
                    width: max-content;
                    background-color: #007bff;
                    color: white;
                    padding: 10px 20px;
                    text-align: center;
                    text-color: white;
                    border-radius: 5px;
                    text-decoration: none;
                    font-weight: bold;
                    margin: 20px auto;
                }
                .footer {
                    text-align: center;
                    font-size: 14px;
                    color: #aaa;
                }
            </style>
        </head>"""

        body = f"""
        <body>
            <div class="container">
                <div class="header">
                    Request Notification
                </div>
                <div class="content">
                    <p>Hello,</p>
                    <p>The status of your recent request has been updated.
                    Below is the latest information regarding it:</p>

                    <div class="request-card">
                        <div class="request-header">Request Details</div>
                        <div class="request-content">

                            <p><strong>ID:</strong> {request_obj.id}</p>
                            <p>
                            <strong>Submitted By:</strong>
                            {request_obj.requesting_user_name} {request_obj.requesting_user_email or ""}
                            </p>
                            <p><strong>Date:</strong> {request_obj.request_time}</p>
                            <div style="display: flex"><p><strong>Status:</strong><div class="badge {badge_color}">{
                                request_obj.status.name
                            }</div></div>
                            <p>
                            <strong>Changes:</strong>
                            {",".join([change.__class__.__name__ for change in request_obj.changes])}
                            </p>
                        </div>
                    </div>
                    <p>If you did not expect this request or have concerns about it,
                    please contact our support team immediately.</p>
                </div>
                <div class="footer">
                    This is an automated message, please do not reply directly to this email. <br>
                    For assistance, please contact our support team.
                </div>
            </div>
        </body>
        """
        return f"""<html>{head} {body}</html>"""



# stdlib
from string import Template

# third party
from result import Err
from result import Ok
from result import Result

# relative
from ...abstract_server import ServerSideType
from ...serde.serializable import serializable
from ...store.document_store import DocumentStore
from ...util.assets import load_png_base64
from ...util.experimental_flags import flags
from ...util.misc_objs import HTMLObject
from ...util.misc_objs import MarkdownDescription
from ...util.notebook_ui.styles import FONT_CSS
from ...util.schema import DO_COMMANDS
from ...util.schema import DS_COMMANDS
from ...util.schema import GUEST_COMMANDS
from ..context import AuthedServiceContext
from ..context import UnauthedServiceContext
from ..notifier.notifier_enums import EMAIL_TYPES
from ..response import SyftError
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import service_method
from ..user.user_roles import ADMIN_ROLE_LEVEL
from ..user.user_roles import GUEST_ROLE_LEVEL
from ..user.user_roles import ServiceRole
from ..warnings import HighSideCRUDWarning
from .settings import ServerSettings
from .settings import ServerSettingsUpdate
from .settings_stash import SettingsStash


@serializable(canonical_name="SettingsService", version=1)
class SettingsService(AbstractService):
    store: DocumentStore
    stash: SettingsStash

    def __init__(self, store: DocumentStore) -> None:
        self.store = store
        self.stash = SettingsStash(store=store)

    @service_method(path="settings.get", name="get")
    def get(self, context: UnauthedServiceContext) -> Result[Ok, Err]:
        """Get Settings"""

        result = self.stash.get_all(context.server.signing_key.verify_key)
        if result.is_ok():
            settings = result.ok()
            # check if the settings list is empty
            if len(settings) == 0:
                return SyftError(message="No settings found")
            result = settings[0]
            return Ok(result)
        else:
            return SyftError(message=result.err())

    @service_method(path="settings.set", name="set")
    def set(
        self, context: AuthedServiceContext, settings: ServerSettings
    ) -> Result[Ok, Err]:
        """Set a new the Server Settings"""
        result = self.stash.set(context.credentials, settings)
        if result.is_ok():
            return result
        else:
            return SyftError(message=result.err())

    @service_method(
        path="settings.update",
        name="update",
        autosplat=["settings"],
        roles=ADMIN_ROLE_LEVEL,
    )
    def update(
        self, context: AuthedServiceContext, settings: ServerSettingsUpdate
    ) -> Result[SyftSuccess, SyftError]:
        res = self._update(context, settings)
        if res.is_ok():
            return SyftSuccess(
                message=(
                    "Settings updated successfully. "
                    + "You must call <client>.refresh() to sync your client with the changes."
                )
            )
        else:
            return SyftError(message=res.err())

    def _update(
        self, context: AuthedServiceContext, settings: ServerSettingsUpdate
    ) -> Result[Ok, Err]:
        """
        Update the Server Settings using the provided values.

        Args:
            name: Optional[str]
                Server name
            organization: Optional[str]
                Organization name
            description: Optional[str]
                Server description
            on_board: Optional[bool]
                Show onboarding panel when a user logs in for the first time
            signup_enabled: Optional[bool]
                Enable/Disable registration
            admin_email: Optional[str]
                Administrator email
            association_request_auto_approval: Optional[bool]

        Returns:
            Result[SyftSuccess, SyftError]: A result indicating the success or failure of the update operation.

        Example:
        >>> server_client.update(name='foo', organization='bar', description='baz', signup_enabled=True)
        SyftSuccess: Settings updated successfully.
        """
        result = self.stash.get_all(context.credentials)
        if result.is_ok():
            current_settings = result.ok()
            if len(current_settings) > 0:
                new_settings = current_settings[0].model_copy(
                    update=settings.to_dict(exclude_empty=True)
                )
                notifier_service = context.server.get_service("notifierservice")

                # If notifications_enabled is present in the update, we need to update the notifier settings
                if settings.notifications_enabled is True:
                    if not notifier_service.settings(context):
                        return SyftError(
                            message="Create notification settings using enable_notifications from user_service"
                        )
                    notifier_service = context.server.get_service("notifierservice")
                    result = notifier_service.set_notifier_active_to_true(context)
                elif settings.notifications_enabled is False:
                    if not notifier_service.settings(context):
                        return SyftError(
                            message="Create notification settings using enable_notifications from user_service"
                        )
                    notifier_service = context.server.get_service("notifierservice")
                    result = notifier_service.set_notifier_active_to_false(context)
                update_result = self.stash.update(context.credentials, new_settings)
                return update_result
            else:
                return Err(value="No settings found")
        else:
            return result

    @service_method(
        path="settings.set_server_side_type_dangerous",
        name="set_server_side_type_dangerous",
        roles=ADMIN_ROLE_LEVEL,
    )
    def set_server_side_type_dangerous(
        self, context: AuthedServiceContext, server_side_type: str
    ) -> Result[SyftSuccess, SyftError]:
        side_type_options = [e.value for e in ServerSideType]
        if server_side_type not in side_type_options:
            return SyftError(
                message=f"Not a valid server_side_type, please use one of the options from: {side_type_options}"
            )

        result = self.stash.get_all(context.credentials)
        if result.is_ok():
            current_settings = result.ok()
            if len(current_settings) > 0:
                new_settings = current_settings[0]
                new_settings.server_side_type = server_side_type
                update_result = self.stash.update(context.credentials, new_settings)
                if update_result.is_ok():
                    return SyftSuccess(
                        message=(
                            "Settings updated successfully. "
                            + "You must call <client>.refresh() to sync your client with the changes."
                        )
                    )
                else:
                    return SyftError(message=update_result.err())
            else:
                return SyftError(message="No settings found")
        else:
            return SyftError(message=result.err())

    @service_method(
        path="settings.enable_notifications",
        name="enable_notifications",
        roles=ADMIN_ROLE_LEVEL,
    )
    def enable_notifications(
        self,
        context: AuthedServiceContext,
        email_username: str | None = None,
        email_password: str | None = None,
        email_sender: str | None = None,
        email_server: str | None = None,
        email_port: str | None = None,
    ) -> SyftSuccess | SyftError:
        notifier_service = context.server.get_service("notifierservice")
        return notifier_service.turn_on(
            context=context,
            email_username=email_username,
            email_password=email_password,
            email_sender=email_sender,
            email_server=email_server,
            email_port=email_port,
        )

    @service_method(
        path="settings.disable_notifications",
        name="disable_notifications",
        roles=ADMIN_ROLE_LEVEL,
    )
    def disable_notifications(
        self,
        context: AuthedServiceContext,
    ) -> SyftSuccess | SyftError:
        notifier_service = context.server.get_service("notifierservice")
        return notifier_service.turn_off(context=context)

    @service_method(
        path="settings.allow_guest_signup",
        name="allow_guest_signup",
        warning=HighSideCRUDWarning(confirmation=True),
    )
    def allow_guest_signup(
        self, context: AuthedServiceContext, enable: bool
    ) -> SyftSuccess | SyftError:
        """Enable/Disable Registration for Data Scientist or Guest Users."""
        flags.CAN_REGISTER = enable

        settings = ServerSettingsUpdate(signup_enabled=enable)
        result = self._update(context=context, settings=settings)

        if isinstance(result, SyftError):
            return SyftError(message=f"Failed to update settings: {result.err()}")

        message = "enabled" if enable else "disabled"
        return SyftSuccess(message=f"Registration feature successfully {message}")

    # NOTE: This service is disabled until we bring back Eager Execution
    # @service_method(
    #     path="settings.enable_eager_execution",
    #     name="enable_eager_execution",
    #     roles=ADMIN_ROLE_LEVEL,
    #     warning=HighSideCRUDWarning(confirmation=True),
    # )
    def enable_eager_execution(
        self, context: AuthedServiceContext, enable: bool
    ) -> SyftSuccess | SyftError:
        """Enable/Disable eager execution."""
        settings = ServerSettingsUpdate(eager_execution_enabled=enable)

        result = self._update(context=context, settings=settings)

        if result.is_err():
            return SyftError(message=f"Failed to update settings: {result.err()}")

        message = "enabled" if enable else "disabled"
        return SyftSuccess(message=f"Eager execution {message}")

    @service_method(path="settings.set_email_rate_limit", name="set_email_rate_limit")
    def set_email_rate_limit(
        self, context: AuthedServiceContext, email_type: EMAIL_TYPES, daily_limit: int
    ) -> SyftSuccess | SyftError:
        notifier_service = context.server.get_service("notifierservice")
        return notifier_service.set_email_rate_limit(context, email_type, daily_limit)

    @service_method(
        path="settings.allow_association_request_auto_approval",
        name="allow_association_request_auto_approval",
    )
    def allow_association_request_auto_approval(
        self, context: AuthedServiceContext, enable: bool
    ) -> SyftSuccess | SyftError:
        new_settings = ServerSettingsUpdate(association_request_auto_approval=enable)
        result = self._update(context, settings=new_settings)
        if isinstance(result, SyftError):
            return result

        message = "enabled" if enable else "disabled"
        return SyftSuccess(
            message="Association request auto-approval successfully " + message
        )

    @service_method(
        path="settings.welcome_preview",
        name="welcome_preview",
    )
    def welcome_preview(
        self,
        context: AuthedServiceContext,
        markdown: str = "",
        html: str = "",
    ) -> MarkdownDescription | HTMLObject | SyftError:
        if not markdown and not html or markdown and html:
            return SyftError(
                message="Invalid markdown/html fields. You must set one of them."
            )

        welcome_msg = None
        if markdown:
            welcome_msg = MarkdownDescription(text=markdown)
        else:
            welcome_msg = HTMLObject(text=html)

        return welcome_msg

    @service_method(
        path="settings.welcome_customize",
        name="welcome_customize",
    )
    def welcome_customize(
        self,
        context: AuthedServiceContext,
        markdown: str = "",
        html: str = "",
    ) -> SyftSuccess | SyftError:
        if not markdown and not html or markdown and html:
            return SyftError(
                message="Invalid markdown/html fields. You must set one of them."
            )

        welcome_msg = None
        if markdown:
            welcome_msg = MarkdownDescription(text=markdown)
        else:
            welcome_msg = HTMLObject(text=html)

        new_settings = ServerSettingsUpdate(welcome_markdown=welcome_msg)
        result = self._update(context=context, settings=new_settings)
        if isinstance(result, SyftError):
            return result

        return SyftSuccess(message="Welcome Markdown was successfully updated!")

    @service_method(
        path="settings.welcome_show",
        name="welcome_show",
        roles=GUEST_ROLE_LEVEL,
    )
    def welcome_show(
        self,
        context: AuthedServiceContext,
    ) -> HTMLObject | MarkdownDescription | SyftError:
        result = self.stash.get_all(context.server.signing_key.verify_key)
        user_service = context.server.get_service("userservice")
        role = user_service.get_role_for_credentials(context.credentials)
        if result.is_ok():
            settings = result.ok()
            # check if the settings list is empty
            if len(settings) == 0:
                return SyftError(message="No settings found")
            settings = settings[0]
            if settings.welcome_markdown:
                str_tmp = Template(settings.welcome_markdown.text)
                welcome_msg_class = type(settings.welcome_markdown)
                server_side_type = (
                    "Low Side"
                    if context.server.metadata.server_side_type
                    == ServerSideType.LOW_SIDE.value
                    else "High Side"
                )
                commands = ""
                if (
                    role.value == ServiceRole.NONE.value
                    or role.value == ServiceRole.GUEST.value
                ):
                    commands = GUEST_COMMANDS
                elif (
                    role is not None and role.value == ServiceRole.DATA_SCIENTIST.value
                ):
                    commands = DS_COMMANDS
                elif role is not None and role.value >= ServiceRole.DATA_OWNER.value:
                    commands = DO_COMMANDS

                command_list = f"""
                <ul style='padding-left: 1em;'>
                    {commands}
                </ul>
                """
                result = str_tmp.safe_substitute(
                    FONT_CSS=FONT_CSS,
                    server_symbol=load_png_base64("small-syft-symbol-logo.png"),
                    datasite_name=context.server.name,
                    description=context.server.metadata.description,
                    # server_url='http://testing:8080',
                    server_type=context.server.metadata.server_type.capitalize(),
                    server_side_type=server_side_type,
                    server_version=context.server.metadata.syft_version,
                    command_list=command_list,
                )
                return welcome_msg_class(text=result)
            return SyftError(message="There's no welcome message")
        else:
            return SyftError(message=result.err())



# stdlib

# third party
from result import Err
from result import Result

# relative
from ...custom_worker.config import DockerWorkerConfig
from ...custom_worker.config import WorkerConfig
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...store.document_store import BaseUIDStoreStash
from ...store.document_store import DocumentStore
from ...store.document_store import PartitionKey
from ...store.document_store import PartitionSettings
from ...store.document_store import QueryKeys
from ..action.action_permissions import ActionObjectPermission
from ..action.action_permissions import ActionPermission
from .worker_image import SyftWorkerImage

WorkerConfigPK = PartitionKey(key="config", type_=WorkerConfig)


@serializable(canonical_name="SyftWorkerImageStash", version=1)
class SyftWorkerImageStash(BaseUIDStoreStash):
    object_type = SyftWorkerImage
    settings: PartitionSettings = PartitionSettings(
        name=SyftWorkerImage.__canonical_name__,
        object_type=SyftWorkerImage,
    )

    def __init__(self, store: DocumentStore) -> None:
        super().__init__(store=store)

    def set(
        self,
        credentials: SyftVerifyKey,
        obj: SyftWorkerImage,
        add_permissions: list[ActionObjectPermission] | None = None,
        add_storage_permission: bool = True,
        ignore_duplicates: bool = False,
    ) -> Result[SyftWorkerImage, str]:
        add_permissions = [] if add_permissions is None else add_permissions

        # By default syft images have all read permission
        add_permissions.append(
            ActionObjectPermission(uid=obj.id, permission=ActionPermission.ALL_READ)
        )

        if isinstance(obj.config, DockerWorkerConfig):
            result = self.get_by_worker_config(
                credentials=credentials, config=obj.config
            )
            if result.is_ok() and result.ok() is not None:
                return Err(f"Image already exists for: {obj.config}")

        return super().set(
            credentials,
            obj,
            add_permissions=add_permissions,
            add_storage_permission=add_storage_permission,
            ignore_duplicates=ignore_duplicates,
        )

    def get_by_worker_config(
        self, credentials: SyftVerifyKey, config: WorkerConfig
    ) -> Result[SyftWorkerImage | None, str]:
        qks = QueryKeys(qks=[WorkerConfigPK.with_obj(config)])
        return self.query_one(credentials=credentials, qks=qks)



# stdlib
import os
import subprocess  # nosec
import sys


def get_commit_hash() -> str:
    cwd = os.path.dirname(os.path.abspath(__file__))
    try:
        output = subprocess.check_output(
            "git rev-parse --short HEAD".split(" "),
            cwd=cwd,  # nosec
        )
        return output.strip().decode("ascii")
    except subprocess.CalledProcessError as e:
        print(f"Error getting commit hash: {e}")
        sys.exit(1)


def update_commit_variable(file_path: str, commit_hash: str) -> None:
    """Replace the __commit__ variable with the actual commit hash."""
    try:
        with open(file_path) as file:
            lines = file.readlines()

        with open(file_path, "w") as file:
            updated = False
            for line in lines:
                if "__commit__ = " in line:
                    file.write(f'__commit__ = "{commit_hash}"\n')
                    updated = True
                else:
                    file.write(line)
            if not updated:
                print("No __commit__ variable found in the file.")
    except OSError as e:
        print(f"Error reading or writing file: {e}")
        sys.exit(1)


if __name__ == "__main__":
    if len(sys.argv) != 2:
        print("Usage: python update_commit.py <file_path>")
        sys.exit(1)

    file_path = sys.argv[1]
    commit_hash = get_commit_hash()
    update_commit_variable(file_path, commit_hash)



# stdlib

# third party

# relative
from ..service.dataset.dataset import Dataset
from ..types.uid import UID
from .client import SyftClient


class SearchResults:
    def __init__(self, results: list[tuple[SyftClient, list[Dataset]]]) -> None:
        self._dataset_client = {}
        self._datasets = []
        for pairs in results:
            client = pairs[0]
            datasets = list(pairs[1])
            for dataset in datasets:
                self._dataset_client[dataset.id] = client
                self._datasets.append(dataset)

    def __getitem__(self, key: int | str | UID) -> Dataset:
        if isinstance(key, int):
            return self._datasets[key]
        else:
            for dataset in self._datasets:
                if isinstance(key, UID):
                    if dataset.id == key:
                        return dataset
                elif isinstance(key, str):
                    if dataset.name == key:
                        return dataset
                    elif str(dataset.id) == key:
                        return dataset
        raise KeyError

    def __repr__(self) -> str:
        return str(self._datasets)

    def _repr_html_(self) -> str:
        return self._datasets._repr_html_()

    def client_for(self, key: Dataset | int | str | UID) -> SyftClient:
        if isinstance(key, Dataset):
            dataset = key
        else:
            dataset = self.__getitem__(key)
        return self._dataset_client[dataset.id]

    def __len__(self) -> int:
        return len(self._datasets)


# class Search:
#     def __init__(self, datasites: DatasiteRegistry) -> None:
#         self.datasites: list[tuple[ServerPeer, ServerMetadataJSON | None]] = (
#             datasites.online_datasites
#         )

#     @staticmethod
#     def __search_one_server(
#         peer_tuple: tuple[ServerPeer, ServerMetadataJSON], name: str
#     ) -> tuple[SyftClient | None, list[Dataset]]:
#         try:
#             peer, server_metadata = peer_tuple
#             client = peer.guest_client
#             results = client.api.services.dataset.search(name=name)
#             return (client, results)
#         except Exception as e:  # noqa
#             warning = SyftWarning(
#                 message=f"Got exception {e} at server {server_metadata.name}"
#             )
#             display(warning)
#             return (None, [])

#     def __search(self, name: str) -> list[tuple[SyftClient, list[Dataset]]]:
#         with ThreadPoolExecutor(max_workers=20) as executor:
#             # results: list[tuple[SyftClient | None, list[Dataset]]] = [
#             #     self.__search_one_server(peer_tuple, name) for peer_tuple in self.datasites
#             # ]
#             results: list[tuple[SyftClient | None, list[Dataset]]] = list(
#                 executor.map(
#                     lambda peer_tuple: self.__search_one_server(peer_tuple, name),
#                     self.datasites,
#                 )
#             )
#         # filter out SyftError
#         filtered = [(client, result) for client, result in results if client and result]

#         return filtered

#     def search(self, name: str) -> SearchResults:
#         """
#         Searches for a specific dataset by name.

#         Args:
#             name (str): The name of the dataset to search for.

#         Returns:
#             SearchResults: An object containing the search results.
#         """
#         return SearchResults(self.__search(name))



# stdlib
from collections.abc import Callable
from collections.abc import Collection
from collections.abc import Iterable
from dataclasses import dataclass
import enum
import html
import logging
import operator
import textwrap
from typing import Any
from typing import ClassVar
from typing import Literal
from typing import TYPE_CHECKING

# third party
import pandas as pd
from rich import box
from rich.console import Console
from rich.console import Group
from rich.markdown import Markdown
from rich.padding import Padding
from rich.panel import Panel
from typing_extensions import Self

# relative
from ...client.api import APIRegistry
from ...client.client import SyftClient
from ...client.sync_decision import SyncDecision
from ...client.sync_decision import SyncDirection
from ...server.credentials import SyftVerifyKey
from ...types.datetime import DateTime
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import SyftObject
from ...types.syft_object import short_uid
from ...types.syncable_object import SyncableSyftObject
from ...types.uid import LineageID
from ...types.uid import UID
from ...util import options
from ...util.colors import SURFACE
from ...util.notebook_ui.components.sync import Label
from ...util.notebook_ui.components.sync import SyncTableObject
from ...util.notebook_ui.icons import Icon
from ...util.notebook_ui.styles import FONT_CSS
from ...util.notebook_ui.styles import ITABLES_CSS
from ..action.action_object import ActionObject
from ..action.action_permissions import ActionObjectPermission
from ..action.action_permissions import ActionPermission
from ..action.action_permissions import StoragePermission
from ..api.api import TwinAPIEndpoint
from ..code.user_code import UserCode
from ..code.user_code import UserCodeStatusCollection
from ..job.job_stash import Job
from ..job.job_stash import JobType
from ..log.log import SyftLog
from ..output.output_service import ExecutionOutput
from ..request.request import Request
from ..response import SyftError
from ..response import SyftSuccess
from ..user.user import UserView
from .sync_state import SyncState

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    # relative
    from .resolve_widget import PaginatedResolveWidget
    from .resolve_widget import ResolveWidget

sketchy_tab = "‎ " * 4


class AttrDiff(SyftObject):
    # version
    __canonical_name__ = "AttrDiff"
    __version__ = SYFT_OBJECT_VERSION_1
    attr_name: str
    low_attr: Any = None
    high_attr: Any = None

    def _repr_html_(self) -> str:
        return f"""{self.attr_name}:
    Low Side value: {self.low_attr}
    High Side value: {self.high_attr}
    """

    def __repr_side__(self, side: str) -> str:
        if side == "low":
            return recursive_attr_repr(self.low_attr)
        else:
            return recursive_attr_repr(self.high_attr)

    def _coll_repr_(self) -> dict[str, Any]:
        return {
            "attr name": self.attr_name,
            "low attr": html.escape(f"{self.low_attr}"),
            "high attr": html.escape(str(self.high_attr)),
        }


class ListDiff(AttrDiff):
    # version
    __canonical_name__ = "ListDiff"
    __version__ = SYFT_OBJECT_VERSION_1
    diff_ids: list[int] = []
    new_low_ids: list[int] = []
    new_high_ids: list[int] = []

    @property
    def is_empty(self) -> bool:
        return (
            len(self.diff_ids) == 0
            and len(self.new_low_ids) == 0
            and len(self.new_high_ids) == 0
        )

    @classmethod
    def from_lists(cls, attr_name: str, low_list: list, high_list: list) -> "ListDiff":
        diff_ids = []
        new_low_ids = []
        new_high_ids = []
        if len(low_list) != len(high_list):
            if len(low_list) > len(high_list):
                common_length = len(high_list)
                new_low_ids = list(range(common_length, len(low_list)))
            else:
                common_length = len(low_list)
                new_high_ids = list(range(common_length, len(high_list)))
        else:
            common_length = len(low_list)

        for i in range(common_length):
            if hasattr(low_list[i], "syft_eq"):
                if not low_list[i].syft_eq(high_list[i]):
                    diff_ids.append(i)
            elif low_list[i] != high_list[i]:
                diff_ids.append(i)

        change_diff = ListDiff(
            attr_name=attr_name,
            low_attr=low_list,
            high_attr=high_list,
            diff_ids=diff_ids,
            new_low_ids=new_low_ids,
            new_high_ids=new_high_ids,
        )
        return change_diff


def recursive_attr_repr(value_attr: list | dict | bytes, num_tabs: int = 0) -> str:
    new_num_tabs = num_tabs + 1

    if isinstance(value_attr, list):
        if len(value_attr) == 1:
            value_attr = value_attr[0]
        else:
            list_repr = "[\n"
            for elem in value_attr:
                list_repr += recursive_attr_repr(elem, num_tabs=num_tabs + 1) + "\n"
            list_repr += "]"
            return list_repr

    elif isinstance(value_attr, dict):
        dict_repr = "{\n"
        for key, elem in value_attr.items():
            dict_repr += f"{sketchy_tab * new_num_tabs}{key}: {str(elem)}\n"
        dict_repr += "}"
        return dict_repr

    elif isinstance(value_attr, bytes):
        value_attr = repr(value_attr)  # type: ignore
        if len(value_attr) > 50:
            value_attr = value_attr[:50] + "..."  # type: ignore

    if isinstance(value_attr, UID):
        value_attr = short_uid(value_attr)  # type: ignore

    return f"{sketchy_tab*num_tabs}{str(value_attr)}"


class ObjectDiff(SyftObject):  # StateTuple (compare 2 objects)
    # version
    __canonical_name__ = "ObjectDiff"
    __version__ = SYFT_OBJECT_VERSION_1
    low_obj: SyncableSyftObject | None = None
    high_obj: SyncableSyftObject | None = None
    low_server_uid: UID
    high_server_uid: UID
    low_permissions: list[str] = []
    high_permissions: list[str] = []
    low_storage_permissions: set[UID] = set()
    high_storage_permissions: set[UID] = set()
    low_status: str | None = None
    high_status: str | None = None
    last_sync_date_low: DateTime | None = None
    last_sync_dat_high: DateTime | None = None

    obj_type: type
    diff_list: list[AttrDiff] = []

    __repr_attrs__ = [
        "low_state",
        "high_state",
    ]
    __syft_include_id_coll_repr__ = False

    def is_mock(self, side: str) -> bool:
        # An object is a mock object if it exists on both sides,
        # and has no storage permissions on `side`
        # NOTE both sides must have the objects, else it is a new object.
        # New+mock objects do not appear naturally, but if they do we
        # want them to show up.
        if side == "low":
            obj = self.low_obj
            other_obj = self.high_obj
            permissions = self.low_storage_permissions
            server_uid = self.low_server_uid
        elif side == "high":
            obj = self.high_obj
            other_obj = self.low_obj
            permissions = self.high_storage_permissions
            server_uid = self.high_server_uid
        else:
            raise ValueError("Invalid side")

        if obj is None or other_obj is None:
            return False

        return server_uid not in permissions

    @classmethod
    def from_objects(
        cls,
        low_obj: SyncableSyftObject | None,
        high_obj: SyncableSyftObject | None,
        low_status: str | None,
        high_status: str | None,
        low_permissions: set[str],
        high_permissions: set[str],
        low_storage_permissions: set[UID],
        high_storage_permissions: set[UID],
        low_server_uid: UID,
        high_server_uid: UID,
        last_sync_date_low: DateTime | None = None,
        last_sync_date_high: DateTime | None = None,
    ) -> "ObjectDiff":
        if low_obj is None and high_obj is None:
            raise ValueError("Both low and high objects are None")
        obj_type = type(low_obj if low_obj is not None else high_obj)

        res = cls(
            low_obj=low_obj,
            high_obj=high_obj,
            low_status=low_status,
            high_status=high_status,
            obj_type=obj_type,
            low_server_uid=low_server_uid,
            high_server_uid=high_server_uid,
            low_permissions=low_permissions,
            high_permissions=high_permissions,
            low_storage_permissions=low_storage_permissions,
            high_storage_permissions=high_storage_permissions,
            last_sync_date_low=last_sync_date_low,
            last_sync_date_high=last_sync_date_high,
        )

        if (
            low_obj is None
            or high_obj is None
            or (res.is_mock("low") and high_status == "SAME")
            or (res.is_mock("high") and low_status == "SAME")
        ):
            diff_list = []
        else:
            diff_list = low_obj.syft_get_diffs(high_obj)

        res.diff_list = diff_list
        return res

    def __hash__(self) -> int:
        return hash(self.object_id) + hash(self.low_obj) + hash(self.high_obj)

    @property
    def last_sync_date(self) -> DateTime | None:
        last_sync_low = self.last_sync_date_low if self.low_obj is not None else None
        last_sync_high = self.last_sync_dat_high if self.high_obj is not None else None

        if last_sync_low is None:
            return last_sync_high
        elif last_sync_high is None:
            return last_sync_low
        else:
            return max(last_sync_low, last_sync_high)

    @property
    def status(self) -> Literal["NEW", "SAME", "MODIFIED"]:
        if self.low_obj is None or self.high_obj is None:
            return "NEW"
        if len(self.diff_list) == 0:
            return "SAME"
        return "MODIFIED"

    @property
    def object_id(self) -> UID:
        uid: UID | LineageID = (
            self.low_obj.id if self.low_obj is not None else self.high_obj.id  # type: ignore
        )
        if isinstance(uid, LineageID):
            return uid.id
        return uid

    @property
    def non_empty_object(self) -> SyftObject | None:
        if self.low_obj is not None:
            return self.low_obj
        else:
            return self.high_obj

    @property
    def object_type(self) -> str:
        return self.obj_type.__name__

    @property
    def high_state(self) -> str:
        return self.state_str("high")

    @property
    def low_state(self) -> str:
        return self.state_str("low")

    @property
    def object_uid(self) -> UID:
        return self.low_obj.id if self.low_obj is not None else self.high_obj.id  # type: ignore

    def repr_attr_diffstatus_dict(self) -> dict:
        # relative
        from .resolve_widget import DiffStatus

        low_attrs = self.repr_attr_dict("low")
        high_attrs = self.repr_attr_dict("high")
        all_attrs = set(low_attrs.keys()) | set(high_attrs.keys())

        res = {}
        for attr in all_attrs:
            value_low = low_attrs.get(attr, None)
            value_high = high_attrs.get(attr, None)

            if value_low is None or value_high is None:
                res[attr] = DiffStatus.NEW
            elif isinstance(value_low, pd.DataFrame) and isinstance(
                value_high, pd.DataFrame
            ):
                res[attr] = (
                    DiffStatus.MODIFIED
                    if not value_low.equals(value_high)
                    else DiffStatus.SAME
                )
            elif value_low != value_high:
                res[attr] = DiffStatus.MODIFIED
            else:
                res[attr] = DiffStatus.SAME
        return res

    def repr_attr_dict(self, side: str) -> dict[str, Any]:
        obj = self.low_obj if side == "low" else self.high_obj
        if isinstance(obj, ActionObject):
            # Only safe for ActionObjects created by data owners
            return {"value": obj.syft_action_data_repr_}
        repr_attrs = getattr(obj, "__repr_attrs__", [])
        res = {}
        for attr in repr_attrs:
            value = getattr(obj, attr)
            res[attr] = value
        return res

    def diff_attributes_str(self, side: str) -> str:
        obj = self.low_obj if side == "low" else self.high_obj

        if obj is None:
            return ""

        repr_attrs = getattr(obj, "__repr_attrs__", [])
        if self.status == "SAME":
            repr_attrs = repr_attrs[:3]

        if self.status in {"SAME", "NEW"}:
            attrs_str = ""
            for attr in repr_attrs:
                value = getattr(obj, attr)
                attrs_str += f"{attr}: {recursive_attr_repr(value)}\n"
            return attrs_str

        elif self.status == "MODIFIED":
            attrs_str = ""
            for diff in self.diff_list:
                attrs_str += f"{diff.attr_name}: {diff.__repr_side__(side)}\n"
            return attrs_str
        else:
            raise ValueError("")

    def diff_side_str(self, side: str) -> str:
        obj = self.low_obj if side == "low" else self.high_obj
        if obj is None:
            return ""
        res = f"{self.obj_type.__name__.upper()} #{short_uid(obj.id)}\n"
        res += self.diff_attributes_str(side)
        return res

    def state_str(self, side: str) -> str:
        other_obj: SyftObject | None = None
        if side == "high":
            obj = self.high_obj
            other_obj = self.low_obj
        else:
            obj = self.low_obj
            other_obj = self.high_obj

        if obj is None:
            return "-"
        if self.status == "SAME":
            return f"SAME\n{self.obj_type.__name__}"

        if isinstance(obj, ActionObject):
            return obj.__repr__()

        if other_obj is None:  # type: ignore[unreachable]
            attrs_str = ""
            attrs = getattr(obj, "__repr_attrs__", [])
            for attr in attrs:
                value = getattr(obj, attr)
                attrs_str += f"{sketchy_tab}{attr} = {recursive_attr_repr(value)}\n"
            attrs_str = attrs_str[:-1]
            return f"NEW\n\nclass {self.object_type}:\n{attrs_str}"

        attr_text = f"DIFF\nclass {self.object_type}:\n"
        for diff in self.diff_list:
            # TODO
            attr_text += (
                f"{sketchy_tab}{diff.attr_name}={diff.__repr_side__(side)}," + "\n"
            )
        if len(self.diff_list) > 0:
            attr_text = attr_text[:-2]

        return attr_text

    def get_obj(self) -> SyftObject | None:
        if self.status == "NEW":
            return self.low_obj if self.low_obj is not None else self.high_obj
        else:
            raise ValueError("Cannot get object from a diff that is not new")

    def _coll_repr_(self) -> dict[str, Any]:
        low_state = f"{self.status}\n{self.diff_side_str('low')}"
        high_state = f"{self.status}\n{self.diff_side_str('high')}"
        return {
            "low_state": html.escape(low_state),
            "high_state": html.escape(high_state),
        }

    def _repr_html_(self) -> str:
        if self.low_obj is None and self.high_obj is None:
            return SyftError(message="Something broke")

        base_str = f"""
        <style>
        {FONT_CSS}
        .syft-dataset {{color: {SURFACE[options.color_theme]};}}
        .syft-dataset h3,
        .syft-dataset p
            {{font-family: 'Open Sans';}}
            {ITABLES_CSS}
        </style>
        <div class='syft-diff'>
        """

        obj_repr: str
        attr_text: str
        if self.low_obj is None:
            if hasattr(self.high_obj, "_repr_html_"):
                obj_repr = self.high_obj._repr_html_()  # type: ignore
            elif hasattr(self.high_obj, "_inner_repr"):
                obj_repr = self.high_obj._inner_repr()  # type: ignore
            else:
                obj_repr = self.__repr__()
            attr_text = (
                f"""
    <h3>{self.object_type} ObjectDiff (New {self.object_type}  on the High Side):</h3>
    """
                + obj_repr
            )

        elif self.high_obj is None:
            if hasattr(self.low_obj, "_repr_html_"):
                obj_repr = self.low_obj._repr_html_()  # type: ignore
            elif hasattr(self.low_obj, "_inner_repr"):
                obj_repr = self.low_obj._inner_repr()  # type: ignore
            else:
                obj_repr = self.__repr__()
            attr_text = (
                f"""
    <h3>{self.object_type} ObjectDiff (New {self.object_type}  on the High Side):</h3>
    """
                + obj_repr
            )

        elif self.status == "SAME":
            obj_repr = "No changes between low side and high side"
        else:
            obj_repr = ""
            for diff in self.diff_list:
                obj_repr += diff.__repr__() + "<br>"

            obj_repr = obj_repr.replace("\n", "<br>")

        attr_text = f"<h3>{self.object_type} ObjectDiff:</h3>\n{obj_repr}"
        return base_str + attr_text

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}[{self.obj_type.__name__}](#{str(self.object_id)})"


def _wrap_text(text: str, width: int, indent: int = 4) -> str:
    """Wrap text, preserving existing line breaks"""
    return "\n".join(
        [
            "\n".join(
                textwrap.wrap(
                    line,
                    width,
                    break_long_words=False,
                    replace_whitespace=False,
                    subsequent_indent=" " * indent,
                )
            )
            for line in text.splitlines()
            if line.strip() != ""
        ]
    )


class ObjectDiffBatch(SyftObject):
    __canonical_name__ = "DiffHierarchy"
    __version__ = SYFT_OBJECT_VERSION_1
    LINE_LENGTH: ClassVar[int] = 100
    INDENT: ClassVar[int] = 4
    ORDER: ClassVar[dict] = {"low": 0, "high": 1}

    __syft_include_id_coll_repr__ = False

    low_server_uid: UID
    high_server_uid: UID
    user_verify_key_low: SyftVerifyKey
    user_verify_key_high: SyftVerifyKey

    # Diffs are ordered in depth-first order,
    # the first diff is the root of the hierarchy
    global_diffs: dict[UID, ObjectDiff]
    global_roots: list[UID]
    global_batches: list["ObjectDiffBatch"] | None = None

    hierarchy_levels: list[int]
    dependencies: dict[UID, list[UID]] = {}
    dependents: dict[UID, list[UID]] = {}
    decision: SyncDecision | None = None
    root_diff: ObjectDiff
    sync_direction: SyncDirection | None

    def resolve(self, build_state: bool = True) -> "ResolveWidget":
        # relative
        from .resolve_widget import ResolveWidget

        return ResolveWidget(self, build_state=build_state)

    def walk_graph(
        self,
        deps: dict[UID, list[UID]],
        include_roots: bool = False,
        include_batch_root: bool = True,
    ) -> list[ObjectDiff]:
        root_id = self.root_diff.object_id
        result = [root_id]
        unvisited = [root_id]
        global_roots = [x for x in self.global_roots if x is not root_id]
        roots = []

        while len(unvisited):
            # Do we update this in the terminal case
            new_servers = []
            for server in unvisited:
                if server in global_roots:
                    roots.append(server)
                else:
                    new_servers += deps.get(server, [])

            new_servers = [n for n in new_servers if n not in result]
            unvisited = new_servers
            result += unvisited

        if include_roots:
            result += roots

        if not include_batch_root:
            result.remove(root_id)

        return [self.global_diffs[r] for r in set(result)]

    @property
    def target_server_uid(self) -> UID:
        if self.sync_direction is None:
            raise ValueError("no direction specified")
        if self.sync_direction == SyncDirection.LOW_TO_HIGH:
            return self.high_server_uid
        else:
            return self.low_server_uid

    @property
    def source_server_uid(self) -> UID:
        if self.sync_direction is None:
            raise ValueError("no direction specified")
        if self.sync_direction == SyncDirection.LOW_TO_HIGH:
            return self.low_server_uid
        else:
            return self.high_server_uid

    @property
    def target_verify_key(self) -> SyftVerifyKey:
        if self.sync_direction is None:
            raise ValueError("no direction specified")
        if self.sync_direction == SyncDirection.LOW_TO_HIGH:
            return self.user_verify_key_high
        else:
            return self.user_verify_key_low

    @property
    def source_verify_key(self) -> SyftVerifyKey:
        if self.sync_direction is None:
            raise ValueError("no direction specified")
        if self.sync_direction == SyncDirection.LOW_TO_HIGH:
            return self.user_verify_key_low
        else:
            return self.user_verify_key_high

    @property
    def source_client(self) -> SyftClient:
        return self.build(self.source_server_uid, self.source_verify_key)

    @property
    def target_client(self) -> SyftClient:
        return self.build(self.target_server_uid, self.target_verify_key)

    def build(self, server_uid: UID, syft_client_verify_key: SyftVerifyKey):  # type: ignore
        # relative
        from ...client.datasite_client import DatasiteClient

        api = APIRegistry.api_for(server_uid, syft_client_verify_key)
        client = DatasiteClient(
            api=api,
            connection=api.connection,  # type: ignore
            credentials=api.signing_key,  # type: ignore
        )
        return client

    def get_dependencies(
        self,
        include_roots: bool = False,
        include_batch_root: bool = True,
    ) -> list[ObjectDiff]:
        return self.walk_graph(
            deps=self.dependencies,
            include_roots=include_roots,
            include_batch_root=include_batch_root,
        )

    @property
    def status(self) -> str:
        if self.root_diff.status == "NEW":
            return "NEW"

        batch_statuses = [
            diff.status for diff in self.get_dependencies(include_roots=False)
        ]
        if all(status == "SAME" for status in batch_statuses):
            return "SAME"

        return "MODIFIED"

    @property
    def is_unchanged(self) -> bool:
        return self.status == "SAME"

    def get_dependents(
        self, include_roots: bool = False, include_batch_root: bool = True
    ) -> list[ObjectDiff]:
        return self.walk_graph(
            deps=self.dependents,
            include_roots=include_roots,
            include_batch_root=include_batch_root,
        )

    def __hash__(self) -> int:
        diffs = self.get_dependents(include_roots=False)
        return sum(hash(x) for x in diffs)

    def ignore(self) -> SyftSuccess | SyftError:
        # relative
        from ...client.syncing import handle_ignore_batch

        return handle_ignore_batch(self, self.global_batches)

    def unignore(self) -> SyftSuccess | SyftError:
        # relative
        from ...client.syncing import handle_unignore_batch

        return handle_unignore_batch(self, self.global_batches)

    @property
    def root_id(self) -> UID:
        return self.root_diff.object_id

    @property
    def root_type(self) -> type:
        return self.root_diff.obj_type

    def decision_badge(self) -> str:
        if self.decision is None:
            return ""
        if self.decision == SyncDecision.IGNORE:
            decision_str = "IGNORED"
            badge_color = "label-red"
        if self.decision == SyncDecision.SKIP:
            decision_str = "SKIPPED"
            badge_color = "label-gray"
        else:
            decision_str = "SYNCED"
            badge_color = "label-green"

        return Label(value=decision_str, label_class=badge_color).to_html()

    @property
    def is_ignored(self) -> bool:
        return self.decision == SyncDecision.IGNORE

    @property
    def is_skipped(self) -> bool:
        return self.decision == SyncDecision.SKIP

    def create_new_resolved_states(
        self,
    ) -> tuple["ResolvedSyncState", "ResolvedSyncState"]:
        """
        Returns new ResolvedSyncState objects for the source and target servers
        """
        resolved_state_low = ResolvedSyncState(
            server_uid=self.low_server_uid, alias="low"
        )
        resolved_state_high = ResolvedSyncState(
            server_uid=self.high_server_uid, alias="high"
        )

        # Return source, target
        if self.sync_direction == SyncDirection.LOW_TO_HIGH:
            return resolved_state_low, resolved_state_high
        else:
            return resolved_state_high, resolved_state_low

    @classmethod
    def from_dependencies(
        cls,
        root_uid: UID,
        obj_dependencies: dict[UID, list[UID]],
        obj_dependents: dict[UID, list[UID]],
        obj_uid_to_diff: dict[UID, ObjectDiff],
        root_ids: list[UID],
        low_server_uid: UID,
        high_server_uid: UID,
        user_verify_key_low: SyftVerifyKey,
        user_verify_key_high: SyftVerifyKey,
        sync_direction: SyncDirection,
    ) -> "ObjectDiffBatch":
        def _build_hierarchy_helper(
            uid: UID, level: int = 0, visited: set | None = None
        ) -> list:
            visited = visited if visited is not None else set()

            if uid in visited:
                return []

            result = [(uid, level)]
            visited.add(uid)
            if uid in obj_dependencies:
                deps = obj_dependencies[uid]
                for dep_uid in obj_dependencies[uid]:
                    if dep_uid not in visited:
                        # NOTE we pass visited + deps to recursive calls, to have
                        # all objects at the highest level in the hierarchy
                        # Example:
                        # ExecutionOutput
                        # -- Job
                        # ---- Result
                        # -- Result
                        # We want to omit Job.Result, because it's already in ExecutionOutput.Result
                        result.extend(
                            _build_hierarchy_helper(
                                uid=dep_uid,
                                level=level + 1,
                                visited=visited | set(deps) - {dep_uid},
                            )
                        )
            return result

        batch_uids = _build_hierarchy_helper(root_uid)
        # levels in the tree that we create
        levels = [level for _, level in batch_uids]

        batch_uids = {uid for uid, _ in batch_uids}  # type: ignore

        return cls(
            global_diffs=obj_uid_to_diff,
            global_roots=root_ids,
            hierarchy_levels=levels,
            dependencies=obj_dependencies,
            dependents=obj_dependents,
            root_diff=obj_uid_to_diff[root_uid],
            low_server_uid=low_server_uid,
            high_server_uid=high_server_uid,
            user_verify_key_low=user_verify_key_low,
            user_verify_key_high=user_verify_key_high,
            sync_direction=sync_direction,
        )

    def flatten_visual_hierarchy(self) -> list[ObjectDiff]:
        def flatten_dict(d: dict) -> list:
            if len(d) == 0:
                return []
            else:
                result = []
                for diff, child in d.items():
                    result.append(diff)
                    result += flatten_dict(child)
                return result

        return flatten_dict(self.get_visual_hierarchy())

    def _repr_html_(self) -> str:
        try:
            diffs = self.flatten_visual_hierarchy()
        except Exception as _:
            return SyftError(
                message=html.escape(
                    "Could not render batch, please use resolve(<batch>) instead."
                )
            )._repr_html_()

        return f"""
<h2> ObjectBatchDiff </h2>
{diffs._repr_html_()}
"""

    def status_badge(self) -> dict[str, str]:
        status = self.status
        if status == "NEW":
            badge_color = "label-green"
        elif status == "SAME":
            badge_color = "label-gray"
        else:
            badge_color = "label-orange"
        return {"value": status.upper(), "type": badge_color}

    def _coll_repr_(self) -> dict[str, Any]:
        no_obj_html = "<p class='diff-state-no-obj'>No object</p>"
        if self.root_diff.low_obj is None:
            low_html = no_obj_html
        else:
            low_html = SyncTableObject(object=self.root_diff.low_obj).to_html()

        if self.root_diff.high_obj is None:
            high_html = no_obj_html
        else:
            high_html = SyncTableObject(object=self.root_diff.high_obj).to_html()

        return {
            "Diff status": self.status_badge(),
            "Public Sync State": low_html,
            "Private sync state": high_html,
            "Decision": self.decision_badge(),
        }

    @property
    def visual_hierarchy(self) -> tuple[type, dict]:
        # Returns
        root_obj = (
            self.root.low_obj if self.root.low_obj is not None else self.root.high_obj
        )
        if isinstance(root_obj, Request):
            return Request, {
                Request: [UserCode],
            }
        elif isinstance(root_obj, UserCode):
            return UserCode, {  # type: ignore
                UserCode: [UserCodeStatusCollection, UserCode],
            }
        elif isinstance(root_obj, Job):
            return UserCode, {  # type: ignore
                UserCode: [ExecutionOutput, UserCode],
                ExecutionOutput: [Job],
                Job: [ActionObject, SyftLog, Job],
            }
        elif isinstance(root_obj, TwinAPIEndpoint):
            return TwinAPIEndpoint, {  # type: ignore
                TwinAPIEndpoint: [],
            }
        else:
            raise ValueError(f"Unknown root type: {self.root.obj_type}")

    @property
    def root(self) -> ObjectDiff:
        return self.root_diff

    def __repr__(self) -> Any:
        try:
            return f"""{self.hierarchy_str('low')}

    {self.hierarchy_str('high')}
    """
        except Exception as _:
            return SyftError(
                message=html.escape(
                    "Could not render batch, please use resolve(<batch>) instead."
                )
            )._repr_html_()

    def _repr_markdown_(self, wrap_as_python: bool = True, indent: int = 0) -> str:
        return ""  # Turns off the _repr_markdown_ of SyftObject

    def _get_visual_hierarchy(
        self, server: ObjectDiff, visited: set[UID] | None = None
    ) -> dict[ObjectDiff, dict]:
        visited = visited if visited is not None else set()
        visited.add(server.object_id)

        _, child_types_map = self.visual_hierarchy
        child_types = child_types_map.get(server.obj_type, [])
        dep_ids = self.dependencies.get(server.object_id, []) + self.dependents.get(
            server.object_id, []
        )

        result = {}
        for child_type in child_types:
            children = [
                n
                for n in self.global_diffs.values()
                if n.object_id in dep_ids
                and isinstance(n.low_obj or n.high_obj, child_type)
            ]
            for child in children:
                if child.object_id not in visited:
                    result[child] = self._get_visual_hierarchy(child, visited=visited)

        return result

    @property
    def visual_root(self) -> ObjectDiff:
        dependecies: list[ObjectDiff] = self.get_dependencies(include_roots=True)
        visual_root_type = self.visual_hierarchy[0]

        visual_roots = [
            diff
            for diff in dependecies
            if isinstance(diff.low_obj or diff.high_obj, visual_root_type)
        ]
        if not len(visual_roots):
            raise ValueError("No visual root found")

        return visual_roots[0]

    @property
    def user_code_high(self) -> UserCode | None:
        """return the user code of the high side of this batch, if it exists"""
        user_code_diff = self.user_code_diff
        if user_code_diff is not None and isinstance(user_code_diff.high_obj, UserCode):
            return user_code_diff.high_obj
        return None

    @property
    def user_code_diff(self) -> ObjectDiff | None:
        """return the main user code diff of the high side of this batch, if it exists"""
        user_code_diffs: list[ObjectDiff] = [
            diff
            for diff in self.get_dependencies(include_roots=True)
            if issubclass(diff.obj_type, UserCode)
        ]

        if len(user_code_diffs) == 0:
            return None
        else:
            # main usercode is always the first, batches are sorted in depth-first order
            return user_code_diffs[0]

    @property
    def user(self) -> UserView | SyftError:
        user_code_diff = self.user_code_diff
        if user_code_diff is not None and isinstance(user_code_diff.low_obj, UserCode):
            return user_code_diff.low_obj.user
        return SyftError(message="No user found")

    def get_visual_hierarchy(self) -> dict[ObjectDiff, dict]:
        visual_root = self.visual_root
        return {visual_root: self._get_visual_hierarchy(self.visual_root)}  # type: ignore

    def _get_obj_str(self, diff_obj: ObjectDiff, level: int, side: str) -> str:
        obj = diff_obj.low_obj if side == "low" else diff_obj.high_obj
        if obj is None:
            return ""
        indent = " " * level * self.INDENT
        obj_str = diff_obj.diff_side_str(side)
        obj_str = _wrap_text(obj_str, width=self.LINE_LENGTH - len(indent))

        line_prefix = indent + f"―――― {diff_obj.status} "
        line = "―" * (self.LINE_LENGTH - len(line_prefix))
        return f"""{line_prefix}{line}

{textwrap.indent(obj_str, indent)}

"""

    def hierarchy_str(self, side: str) -> str:
        def _hierarchy_str_recursive(tree: dict, level: int) -> str:
            result = ""
            for server, children in tree.items():
                result += self._get_obj_str(server, level, side)
                result += _hierarchy_str_recursive(children, level + 1)
            return result

        visual_hierarchy = self.get_visual_hierarchy()
        res = _hierarchy_str_recursive(visual_hierarchy, 0)
        if res == "":
            res = f"No {side} side changes."
        return f"""{side.upper()} SIDE STATE:

{res}"""


class IgnoredBatchView(SyftObject):
    __canonical_name__ = "IgnoredBatchView"
    __version__ = SYFT_OBJECT_VERSION_1
    batch: ObjectDiffBatch
    other_batches: list[ObjectDiffBatch]

    def _coll_repr_(self) -> dict[str, Any]:
        return self.batch._coll_repr_()

    def _repr_html_(self) -> str:
        return self.batch._repr_html_()

    def stage_change(self) -> None:
        self.batch.decision = None
        required_dependencies = {
            d.object_id for d in self.batch.get_dependencies(include_roots=True)
        }

        for other_batch in self.other_batches:
            if (
                other_batch.decision == SyncDecision.IGNORE
                and other_batch.root_id in required_dependencies
            ):
                logger.debug(f"ignoring other batch ({other_batch.root_type.__name__})")
                other_batch.decision = None


class FilterProperty(enum.Enum):
    USER = enum.auto()
    TYPE = enum.auto()
    STATUS = enum.auto()
    IGNORED = enum.auto()

    def from_batch(self, batch: ObjectDiffBatch) -> Any:
        if self == FilterProperty.USER:
            user = batch.user
            if isinstance(user, UserView):
                return user.email
            return None
        elif self == FilterProperty.TYPE:
            return batch.root_diff.obj_type.__name__.lower()
        elif self == FilterProperty.STATUS:
            return batch.status.lower()
        elif self == FilterProperty.IGNORED:
            return batch.is_ignored
        else:
            raise ValueError(f"Invalid property: {property}")


@dataclass
class ServerDiffFilter:
    """
    Filter to apply to a ServerDiff object to determine if it should be included in a batch.

    Checks for `property op value` , where
        property: FilterProperty - property to filter on
        value: Any - value to compare against
        op: callable[[Any, Any], bool] - comparison operator. Default is `operator.eq`

    If the comparison fails, the batch is excluded.
    """

    filter_property: FilterProperty
    filter_value: Any
    op: Callable[[Any, Any], bool] = operator.eq

    def __call__(self, batch: ObjectDiffBatch) -> bool:
        filter_value = self.filter_value
        if isinstance(filter_value, str):
            filter_value = filter_value.lower()

        try:
            p = self.filter_property.from_batch(batch)
            if self.op == operator.contains:
                # Contains check has reversed arg order: check if p in self.filter_value
                return p in filter_value
            else:
                return self.op(p, filter_value)
        except Exception as e:
            # By default, exclude the batch if there is an error
            logger.debug(f"Error filtering batch {batch} with {self}: {e}")
            return False


class ServerDiff(SyftObject):
    __canonical_name__ = "ServerDiff"
    __version__ = SYFT_OBJECT_VERSION_1

    low_server_uid: UID
    high_server_uid: UID
    user_verify_key_low: SyftVerifyKey
    user_verify_key_high: SyftVerifyKey
    obj_uid_to_diff: dict[UID, ObjectDiff] = {}
    obj_dependencies: dict[UID, list[UID]] = {}
    batches: list[ObjectDiffBatch] = []
    all_batches: list[ObjectDiffBatch] = []
    low_state: SyncState
    high_state: SyncState
    direction: SyncDirection | None
    filters: list[ServerDiffFilter] = []

    include_ignored: bool = False

    def resolve(
        self, build_state: bool = True
    ) -> "PaginatedResolveWidget | SyftSuccess":
        if len(self.batches) == 0:
            return SyftSuccess(message="No batches to resolve")

        # relative
        from .resolve_widget import PaginatedResolveWidget

        return PaginatedResolveWidget(batches=self.batches, build_state=build_state)

    def __getitem__(self, idx: Any) -> ObjectDiffBatch:
        return self.batches[idx]

    @property
    def ignored_batches(self) -> list[ObjectDiffBatch]:
        return [
            batch for batch in self.all_batches if batch.decision == SyncDecision.IGNORE
        ]

    @property
    def active_batches(self) -> Iterable[ObjectDiffBatch]:
        decisions_to_skip = {SyncDecision.IGNORE, SyncDecision.SKIP}
        # self.batches might be modified during iteration
        for batch in self.batches:
            if batch.decision not in decisions_to_skip:
                yield batch

    @property
    def ignored_changes(self) -> list[IgnoredBatchView]:
        result = []
        for ignored_batch in self.ignored_batches:
            other_batches = [b for b in self.all_batches if b is not ignored_batch]
            result.append(
                IgnoredBatchView(batch=ignored_batch, other_batches=other_batches)
            )
        return result

    @classmethod
    def from_sync_state(
        cls: type["ServerDiff"],
        low_state: SyncState,
        high_state: SyncState,
        direction: SyncDirection,
        include_ignored: bool = False,
        include_same: bool = False,
        filter_by_email: str | None = None,
        include_types: Collection[type | str] | None = None,
        exclude_types: Collection[type | str] | None = None,
        _include_server_status: bool = False,
    ) -> "ServerDiff":
        obj_uid_to_diff = {}
        for obj_id in set(low_state.objects.keys()) | set(high_state.objects.keys()):
            low_obj = low_state.objects.get(obj_id, None)
            high_obj = high_state.objects.get(obj_id, None)

            low_permissions = low_state.permissions.get(obj_id, set())
            high_permissions = high_state.permissions.get(obj_id, set())

            low_storage_permissions = low_state.storage_permissions.get(obj_id, set())
            high_storage_permissions = high_state.storage_permissions.get(obj_id, set())

            last_sync_date_low = low_state.object_sync_dates.get(obj_id, None)
            last_sync_date_high = high_state.object_sync_dates.get(obj_id, None)

            if _include_server_status:
                low_status = low_state.get_status(obj_id)
                high_status = high_state.get_status(obj_id)
            else:
                low_status = "NEW"
                high_status = "NEW"

            # We don't support deletion of objects yet.
            # So, skip if the object is not present on the *source* side
            source_obj = low_obj if direction == SyncDirection.LOW_TO_HIGH else high_obj
            if source_obj is None:
                continue

            diff = ObjectDiff.from_objects(
                low_obj=low_obj,
                high_obj=high_obj,
                low_status=low_status,
                high_status=high_status,
                low_permissions=low_permissions,
                high_permissions=high_permissions,
                low_storage_permissions=low_storage_permissions,
                high_storage_permissions=high_storage_permissions,
                low_server_uid=low_state.server_uid,
                high_server_uid=high_state.server_uid,
                last_sync_date_low=last_sync_date_low,
                last_sync_date_high=last_sync_date_high,
            )
            obj_uid_to_diff[diff.object_id] = diff

        # TODO move static methods to ServerDiff __init__
        obj_dependencies = ServerDiff.dependencies_from_states(low_state, high_state)
        all_batches = ServerDiff._create_batches(
            low_state,
            high_state,
            obj_dependencies,
            obj_uid_to_diff,
            direction=direction,
        )

        # TODO: Check if high and low ignored batches are the same else error
        previously_ignored_batches = low_state.ignored_batches
        ServerDiff.apply_previous_ignore_state(all_batches, previously_ignored_batches)

        res = cls(
            low_server_uid=low_state.server_uid,
            high_server_uid=high_state.server_uid,
            user_verify_key_low=low_state.syft_client_verify_key,
            user_verify_key_high=high_state.syft_client_verify_key,
            obj_uid_to_diff=obj_uid_to_diff,
            obj_dependencies=obj_dependencies,
            batches=all_batches,
            all_batches=all_batches,
            low_state=low_state,
            high_state=high_state,
            direction=direction,
            filters=[],
        )

        res._filter(
            user_email=filter_by_email,
            include_types=include_types,
            include_ignored=include_ignored,
            include_same=include_same,
            exclude_types=exclude_types,
            inplace=True,
        )

        return res

    @staticmethod
    def apply_previous_ignore_state(
        batches: list[ObjectDiffBatch], previously_ignored_batches: dict[UID, int]
    ) -> None:
        """
        Loop through all ignored batches in syncstate. If batch did not change, set to ignored
        If another batch needs to exist in order to accept that changed batch: also unignore
        e.g. if a job changed, also unignore the usercode
        """

        for root_id, batch_hash in previously_ignored_batches.items():
            for batch in batches:
                if batch.root_id == root_id:
                    if hash(batch) == batch_hash:
                        batch.decision = SyncDecision.IGNORE
                    else:
                        logger.debug(
                            f"""A batch with type {batch.root_type.__name__} was previously ignored but has changed
It will be available for review again."""
                        )
                        # batch has changed, so unignore
                        batch.decision = None
                        # then we also set the dependent batches to unignore
                        # currently we dont do this recusively
                        required_dependencies = {
                            d.object_id
                            for d in batch.get_dependencies(include_roots=True)
                        }

                        for other_batch in batches:
                            if other_batch is not batch:
                                other_batch_root_id = {other_batch.root_id}
                                # if there is overlap
                                if len(required_dependencies & other_batch_root_id):
                                    other_batch.decision = None

    @staticmethod
    def dependencies_from_states(
        low_state: SyncState, high_state: SyncState
    ) -> dict[UID, list[UID]]:
        dependencies = {}
        all_parents = set(low_state.dependencies.keys()) | set(
            high_state.dependencies.keys()
        )
        for parent in all_parents:
            low_deps = low_state.dependencies.get(parent, [])
            high_deps = high_state.dependencies.get(parent, [])
            dependencies[parent] = list(set(low_deps) | set(high_deps))
        return dependencies

    @property
    def diffs(self) -> list[ObjectDiff]:
        diffs_depthfirst = [
            diff
            for hierarchy in self.batches
            for diff in hierarchy.get_dependents(include_roots=False)
        ]
        # deduplicate
        diffs = []
        ids = set()
        for diff in diffs_depthfirst:
            if diff.object_id not in ids:
                diffs.append(diff)
                ids.add(diff.object_id)
        return diffs

    def _repr_markdown_(self) -> None:  # type: ignore
        return None

    def _repr_html_(self) -> Any:
        n = len(self.batches)
        if self.direction == SyncDirection.LOW_TO_HIGH:
            name1 = "Low-side Server"
            name2 = "High-side Server"
        else:
            name1 = "High-side Server"
            name2 = "Low-side Server"
        repr_html = f"""
        <p style="margin-bottom:16px;"></p>
        <div class="diff-state-intro">Comparing sync states</div>
        <p style="margin-bottom:16px;"></p>
        <div class="diff-state-header"><span>{name1}</span> {Icon.ARROW.svg} <span>{name2}</span></div>
        <p style="margin-bottom:16px;"></p>
        <div class="diff-state-sub-header"> This would sync <span class="diff-state-orange-text">{n} batches</span> from <i>{name1}</i> to <i>{name2}</i></div>
        """  # noqa: E501
        repr_html = repr_html.replace("\n", "")

        res = repr_html + self.batches._repr_html_()
        return res

    @staticmethod
    def _sort_batches(hierarchies: list[ObjectDiffBatch]) -> list[ObjectDiffBatch]:
        without_usercode = []
        grouped_by_usercode: dict[UID, list[ObjectDiffBatch]] = {}
        for hierarchy in hierarchies:
            has_usercode = False
            for diff in hierarchy.get_dependencies(include_roots=True):
                obj = diff.low_obj if diff.low_obj is not None else diff.high_obj
                if isinstance(obj, UserCode):
                    usercode_id = obj.id
                    if usercode_id not in grouped_by_usercode:
                        grouped_by_usercode[usercode_id] = []
                    grouped_by_usercode[usercode_id].append(hierarchy)
                    has_usercode = True
                    break
            if not has_usercode:
                without_usercode.append(hierarchy)

        # Order of hierarchies, by root object type
        hierarchy_order = [UserCode, Request, Job]
        # Sort group by hierarchy_order, then by root object id
        for hierarchy_group in grouped_by_usercode.values():
            hierarchy_group.sort(
                key=lambda x: (
                    hierarchy_order.index(x.root.obj_type),
                    x.root.object_id,
                )
            )

        # sorted = sorted groups + without_usercode
        sorted_hierarchies = []
        for grp in grouped_by_usercode.values():
            sorted_hierarchies.extend(grp)
        sorted_hierarchies.extend(without_usercode)
        return sorted_hierarchies

    @staticmethod
    def _create_batches(
        low_sync_state: SyncState,
        high_sync_state: SyncState,
        obj_dependencies: dict[UID, list[UID]],
        obj_uid_to_diff: dict[UID, ObjectDiff],
        direction: SyncDirection,
    ) -> list[ObjectDiffBatch]:
        batches: list[ObjectDiffBatch] = []
        root_ids = []

        for diff in obj_uid_to_diff.values():
            diff_obj = diff.low_obj if diff.low_obj is not None else diff.high_obj
            if isinstance(diff_obj, Request | UserCode | TwinAPIEndpoint):
                # TODO: Figure out nested user codes, do we even need that?

                root_ids.append(diff.object_id)  # type: ignore
            elif (  # type: ignore[unreachable]
                isinstance(diff_obj, Job)  # type: ignore
                and diff_obj.parent_job_id is None
                # ignore Job objects created by TwinAPIEndpoint
                and diff_obj.job_type != JobType.TWINAPIJOB
            ):
                root_ids.append(diff.object_id)  # type: ignore

        # Dependents are the reverse edges of the dependency graph
        obj_dependents = {}
        for parent, children in obj_dependencies.items():
            for child in children:
                obj_dependents[child] = obj_dependencies.get(child, []) + [parent]

        for root_uid in root_ids:
            batch = ObjectDiffBatch.from_dependencies(
                root_uid,
                obj_dependencies,
                obj_dependents,
                obj_uid_to_diff,
                root_ids,
                low_sync_state.server_uid,
                high_sync_state.server_uid,
                low_sync_state.syft_client_verify_key,
                high_sync_state.syft_client_verify_key,
                sync_direction=direction,
            )
            batches.append(batch)

        # TODO ref back to ServerDiff would clean up a lot of logic,
        # No need to save ServerDiff state on every batch
        for batch in batches:
            batch.global_batches = batches

        hierarchies_sorted = ServerDiff._sort_batches(batches)
        return hierarchies_sorted

    @property
    def is_same(self) -> bool:
        return all(object_diff.status == "SAME" for object_diff in self.diffs)

    def _apply_filters(
        self, filters: list[ServerDiffFilter], inplace: bool = True
    ) -> Self:
        """
        Apply filters to the ServerDiff object and return a new ServerDiff object
        """
        batches = self.all_batches
        for filter in filters:
            batches = [b for b in batches if filter(b)]

        if inplace:
            self.filters = filters
            self.batches = batches
            return self
        else:
            return ServerDiff(
                low_server_uid=self.low_server_uid,
                high_server_uid=self.high_server_uid,
                user_verify_key_low=self.user_verify_key_low,
                user_verify_key_high=self.user_verify_key_high,
                obj_uid_to_diff=self.obj_uid_to_diff,
                obj_dependencies=self.obj_dependencies,
                batches=batches,
                all_batches=self.all_batches,
                low_state=self.low_state,
                high_state=self.high_state,
                direction=self.direction,
                filters=filters,
            )

    def _filter(
        self,
        user_email: str | None = None,
        include_ignored: bool = False,
        include_same: bool = False,
        include_types: Collection[str | type] | None = None,
        exclude_types: Collection[type | str] | None = None,
        inplace: bool = True,
    ) -> Self:
        new_filters = []
        if user_email is not None:
            new_filters.append(
                ServerDiffFilter(FilterProperty.USER, user_email, operator.eq)
            )
        if not include_ignored:
            new_filters.append(
                ServerDiffFilter(FilterProperty.IGNORED, True, operator.ne)
            )
        if not include_same:
            new_filters.append(
                ServerDiffFilter(FilterProperty.STATUS, "SAME", operator.ne)
            )
        if include_types is not None:
            include_types_ = {
                t.__name__ if isinstance(t, type) else t for t in include_types
            }
            new_filters.append(
                ServerDiffFilter(FilterProperty.TYPE, include_types_, operator.contains)
            )
        if exclude_types:
            for exclude_type in exclude_types:
                if isinstance(exclude_type, type):
                    exclude_type = exclude_type.__name__
                new_filters.append(
                    ServerDiffFilter(FilterProperty.TYPE, exclude_type, operator.ne)
                )

        return self._apply_filters(new_filters, inplace=inplace)


class SyncInstruction(SyftObject):
    __canonical_name__ = "SyncDecision"
    __version__ = SYFT_OBJECT_VERSION_1

    diff: ObjectDiff
    decision: SyncDecision | None
    new_permissions_lowside: list[ActionObjectPermission]
    new_storage_permissions_lowside: list[StoragePermission]
    new_storage_permissions_highside: list[StoragePermission]
    unignore: bool = False
    mockify: bool

    @classmethod
    def from_batch_decision(
        cls,
        diff: ObjectDiff,
        sync_direction: SyncDirection,
        share_private_data: bool,
        mockify: bool,
        decision: SyncDecision,
        share_to_user: SyftVerifyKey | None,
    ) -> Self:
        # read widget state
        new_permissions_low_side = []

        # read permissions
        if sync_direction == SyncDirection.HIGH_TO_LOW:
            # To create read permissions for the object
            # job/usercode/request/TwinAPIEndpoint
            if share_private_data:  # or diff.object_type == "Job":
                if share_to_user is None:
                    # job ran by another user
                    if (
                        diff.object_type != "Job"
                        and diff.object_type != "ExecutionOutput"
                    ):
                        raise ValueError(
                            "share_to_user is required to share private data"
                        )
                else:
                    new_permissions_low_side = [
                        ActionObjectPermission(
                            uid=diff.object_id,
                            permission=ActionPermission.READ,
                            credentials=share_to_user,
                        )
                    ]

        # storage permissions
        new_storage_permissions = []

        if sync_direction == SyncDirection.HIGH_TO_LOW:
            # TODO: apply storage permissions on both ends
            if not mockify:
                new_storage_permissions.append(
                    StoragePermission(
                        uid=diff.object_id, server_uid=diff.low_server_uid
                    )
                )
        elif sync_direction == SyncDirection.LOW_TO_HIGH:
            new_storage_permissions.append(
                StoragePermission(uid=diff.object_id, server_uid=diff.high_server_uid)
            )

        return cls(
            diff=diff,
            decision=decision,
            new_permissions_lowside=new_permissions_low_side,
            new_storage_permissions_lowside=new_storage_permissions,
            new_storage_permissions_highside=new_storage_permissions,
            mockify=mockify,
        )


class ResolvedSyncState(SyftObject):
    __canonical_name__ = "SyncUpdate"
    __version__ = SYFT_OBJECT_VERSION_1

    server_uid: UID
    create_objs: list[SyncableSyftObject] = []
    update_objs: list[SyncableSyftObject] = []
    delete_objs: list[SyftObject] = []
    new_permissions: list[ActionObjectPermission] = []
    new_storage_permissions: list[StoragePermission] = []
    ignored_batches: dict[UID, int] = {}  # batch root uid -> hash of the batch
    unignored_batches: set[UID] = set()
    alias: str

    @classmethod
    def from_client(cls, client: SyftClient) -> "ResolvedSyncState":
        alias: str = client.metadata.server_side_type  # type: ignore
        if alias not in ["low", "high"]:
            raise ValueError(
                "can only create resolved sync state for high, low side deployments"
            )
        return cls(server_uid=client.id, alias=alias)

    def add_ignored(self, batch: ObjectDiffBatch) -> None:
        self.ignored_batches[batch.root_id] = hash(batch)

    def add_unignored(self, root_id: UID) -> None:
        self.unignored_batches.add(root_id)

    def add_sync_instruction(self, sync_instruction: SyncInstruction) -> None:
        if (
            sync_instruction.decision == SyncDecision.IGNORE
            or sync_instruction.decision == SyncDecision.SKIP
        ):
            return
        diff = sync_instruction.diff

        if sync_instruction.unignore:
            self.unignored_batches.add(sync_instruction.batch_diff.root_id)

        if diff.status == "SAME":
            return

        my_obj = diff.low_obj if self.alias == "low" else diff.high_obj
        other_obj = diff.low_obj if self.alias == "high" else diff.high_obj

        if other_obj is not None and sync_instruction.mockify:
            other_obj = other_obj.create_shareable_sync_copy(mock=True)

        if (
            sync_instruction.decision and sync_instruction.decision.value != self.alias
        ):  # chose for the other
            if diff.status == "MODIFIED":
                # keep IDs comparison here, otherwise it will break with actionobjects
                if other_obj.id not in [x.id for x in self.update_objs]:  # type: ignore
                    self.update_objs.append(other_obj)

            elif diff.status == "NEW":
                if my_obj is None:
                    # keep IDs comparison here, otherwise it will break with actionobjects
                    if other_obj.id not in [x.id for x in self.create_objs]:  # type: ignore
                        self.create_objs.append(other_obj)

                elif other_obj is None:
                    # keep IDs comparison here, otherwise it will break with actionobjects
                    if my_obj.id not in [x.id for x in self.delete_objs]:
                        self.delete_objs.append(my_obj)

        if self.alias == "low":
            self.new_permissions.extend(sync_instruction.new_permissions_lowside)
            self.new_storage_permissions.extend(
                sync_instruction.new_storage_permissions_lowside
            )
        elif self.alias == "high":
            self.new_storage_permissions.extend(
                sync_instruction.new_storage_permissions_highside
            )
        else:
            raise ValueError("Invalid alias")

    @property
    def is_empty(self) -> bool:
        return (
            len(self.create_objs) == 0
            and len(self.update_objs) == 0
            and len(self.delete_objs) == 0
            and len(self.new_permissions) == 0
            and len(self.new_storage_permissions) == 0
        )

    def __repr__(self) -> str:
        return (
            f"ResolvedSyncState(\n"
            f"  create_objs={self.create_objs},\n"
            f"  update_objs={self.update_objs},\n"
            f"  delete_objs={self.delete_objs}\n"
            f"  new_permissions={self.new_permissions}\n"
            f"  new_storage_permissions={self.new_storage_permissions}\n"
            f"  ignored_batches={list(self.ignored_batches.keys())}\n"
            f")"
        )


def display_diff_object(obj_state: str | None) -> Panel:
    if obj_state is None:
        return Panel(Markdown("None"), box=box.ROUNDED, expand=False)
    return Panel(
        Markdown(f"```python\n{obj_state}\n```", code_theme="default"),
        box=box.ROUNDED,
        expand=False,
    )


def display_diff_hierarchy(diff_hierarchy: list[tuple[ObjectDiff, int]]) -> None:
    console = Console()

    for diff, level in diff_hierarchy:
        title = f"{diff.obj_type.__name__}({diff.object_id}) - State: {diff.status}"

        low_side_panel = display_diff_object(diff.low_state if diff.low_obj else None)
        low_side_panel.title = "Low side"
        low_side_panel.title_align = "left"
        high_side_panel = display_diff_object(
            diff.high_state if diff.high_obj is not None else None
        )
        high_side_panel.title = "High side"
        high_side_panel.title_align = "left"

        grouped_panels = Group(low_side_panel, high_side_panel)

        diff_panel = Panel(
            grouped_panels,
            title=title,
            title_align="left",
            box=box.ROUNDED,
            expand=False,
            padding=(1, 2),
        )

        if level > 0:
            diff_panel = Padding(diff_panel, (0, 0, 0, 5 * level))

        console.print(diff_panel)



# stdlib
import base64
import binascii
from collections.abc import AsyncGenerator
import logging
from typing import Annotated

# third party
from fastapi import APIRouter
from fastapi import Body
from fastapi import Depends
from fastapi import HTTPException
from fastapi import Request
from fastapi import Response
from fastapi.responses import JSONResponse
from fastapi.responses import StreamingResponse
from pydantic import ValidationError
import requests

# relative
from ..abstract_server import AbstractServer
from ..client.connection import ServerConnection
from ..protocol.data_protocol import PROTOCOL_TYPE
from ..serde.deserialize import _deserialize as deserialize
from ..serde.serialize import _serialize as serialize
from ..service.context import ServerServiceContext
from ..service.context import UnauthedServiceContext
from ..service.metadata.server_metadata import ServerMetadataJSON
from ..service.response import SyftError
from ..service.user.user import UserCreate
from ..service.user.user import UserPrivateKey
from ..service.user.user_service import UserService
from ..types.uid import UID
from ..util.telemetry import TRACE_MODE
from .credentials import SyftVerifyKey
from .credentials import UserLoginCredentials
from .worker import Worker

logger = logging.getLogger(__name__)


def make_routes(worker: Worker) -> APIRouter:
    if TRACE_MODE:
        # third party
        try:
            # third party
            from opentelemetry import trace
            from opentelemetry.propagate import extract
        except Exception as e:
            logger.error("Failed to import opentelemetry", exc_info=e)

    router = APIRouter()

    async def get_body(request: Request) -> bytes:
        return await request.body()

    def _get_server_connection(peer_uid: UID) -> ServerConnection:
        # relative
        from ..service.network.server_peer import route_to_connection

        network_service = worker.get_service("NetworkService")
        peer = network_service.stash.get_by_uid(worker.verify_key, peer_uid).ok()
        peer_server_route = peer.pick_highest_priority_route()
        connection = route_to_connection(route=peer_server_route)
        return connection

    @router.get("/stream/{peer_uid}/{url_path}/", name="stream")
    async def stream_download(peer_uid: str, url_path: str) -> StreamingResponse:
        try:
            url_path_parsed = base64.urlsafe_b64decode(url_path.encode()).decode()
        except binascii.Error:
            raise HTTPException(404, "Invalid `url_path`.")

        peer_uid_parsed = UID.from_string(peer_uid)

        try:
            peer_connection = _get_server_connection(peer_uid_parsed)
            url = peer_connection.to_blob_route(url_path_parsed)
            stream_response = peer_connection._make_get(url.path, stream=True)
        except requests.RequestException:
            raise HTTPException(404, "Failed to retrieve data from datasite.")

        return StreamingResponse(stream_response, media_type="text/event-stream")

    async def read_request_body_in_chunks(
        request: Request,
    ) -> AsyncGenerator[bytes, None]:
        async for chunk in request.stream():
            yield chunk

    @router.put("/stream/{peer_uid}/{url_path}/", name="stream")
    async def stream_upload(peer_uid: str, url_path: str, request: Request) -> Response:
        try:
            url_path_parsed = base64.urlsafe_b64decode(url_path.encode()).decode()
        except binascii.Error:
            raise HTTPException(404, "Invalid `url_path`.")

        data = await request.body()

        peer_uid_parsed = UID.from_string(peer_uid)

        try:
            peer_connection = _get_server_connection(peer_uid_parsed)
            url = peer_connection.to_blob_route(url_path_parsed)

            print("Url on stream", url.path)
            response = peer_connection._make_put(url.path, data=data, stream=True)
        except requests.RequestException:
            raise HTTPException(404, "Failed to upload data to datasite")

        return Response(
            content=response.content,
            headers=response.headers,
            media_type="application/octet-stream",
        )

    @router.get(
        "/",
        name="healthcheck",
        status_code=200,
        response_class=JSONResponse,
    )
    def root() -> dict[str, str]:
        """
        Currently, all service backends must satisfy either of the following requirements to
        pass the HTTP health checks sent to it from the GCE loadbalancer: 1. Respond with a
        200 on '/'. The content does not matter. 2. Expose an arbitrary url as a readiness
        probe on the pods backing the Service.
        """
        return {"status": "ok"}

    # provide information about the server in JSON
    @router.get("/metadata", response_class=JSONResponse)
    def syft_metadata() -> JSONResponse:
        return worker.metadata.to(ServerMetadataJSON)

    @router.get("/metadata_capnp")
    def syft_metadata_capnp() -> Response:
        result = worker.metadata
        return Response(
            serialize(result, to_bytes=True),
            media_type="application/octet-stream",
        )

    def handle_syft_new_api(
        user_verify_key: SyftVerifyKey, communication_protocol: PROTOCOL_TYPE
    ) -> Response:
        return Response(
            serialize(
                worker.get_api(user_verify_key, communication_protocol), to_bytes=True
            ),
            media_type="application/octet-stream",
        )

    # get the SyftAPI object
    @router.get("/api")
    def syft_new_api(
        request: Request, verify_key: str, communication_protocol: PROTOCOL_TYPE
    ) -> Response:
        user_verify_key: SyftVerifyKey = SyftVerifyKey.from_string(verify_key)
        if TRACE_MODE:
            with trace.get_tracer(syft_new_api.__module__).start_as_current_span(
                syft_new_api.__qualname__,
                context=extract(request.headers),
                kind=trace.SpanKind.SERVER,
            ):
                return handle_syft_new_api(user_verify_key, communication_protocol)
        else:
            return handle_syft_new_api(user_verify_key, communication_protocol)

    def handle_new_api_call(data: bytes) -> Response:
        obj_msg = deserialize(blob=data, from_bytes=True)
        result = worker.handle_api_call(api_call=obj_msg)
        return Response(
            serialize(result, to_bytes=True),
            media_type="application/octet-stream",
        )

    # make a request to the SyftAPI
    @router.post("/api_call")
    def syft_new_api_call(
        request: Request, data: Annotated[bytes, Depends(get_body)]
    ) -> Response:
        if TRACE_MODE:
            with trace.get_tracer(syft_new_api_call.__module__).start_as_current_span(
                syft_new_api_call.__qualname__,
                context=extract(request.headers),
                kind=trace.SpanKind.SERVER,
            ):
                return handle_new_api_call(data)
        else:
            return handle_new_api_call(data)

    def handle_login(email: str, password: str, server: AbstractServer) -> Response:
        try:
            login_credentials = UserLoginCredentials(email=email, password=password)
        except ValidationError as e:
            return {"Error": e.json()}

        method = server.get_service_method(UserService.exchange_credentials)
        context = UnauthedServiceContext(
            server=server, login_credentials=login_credentials
        )
        result = method(context=context)

        if isinstance(result, SyftError):
            logger.error(f"Login Error: {result.message}. user={email}")
            response = result
        else:
            user_private_key = result
            if not isinstance(user_private_key, UserPrivateKey):
                raise Exception(f"Incorrect return type: {type(user_private_key)}")
            response = user_private_key

        return Response(
            serialize(response, to_bytes=True),
            media_type="application/octet-stream",
        )

    def handle_register(data: bytes, server: AbstractServer) -> Response:
        user_create = deserialize(data, from_bytes=True)

        if not isinstance(user_create, UserCreate):
            raise Exception(f"Incorrect type received: {user_create}")

        context = ServerServiceContext(server=server)
        method = server.get_method_with_context(UserService.register, context)

        result = method(new_user=user_create)

        if isinstance(result, SyftError):
            logger.error(
                f"Register Error: {result.message}. user={user_create.model_dump()}"
            )
            response = SyftError(message=f"{result.message}")
        else:
            response = result

        return Response(
            serialize(response, to_bytes=True),
            media_type="application/octet-stream",
        )

    # exchange email and password for a SyftSigningKey
    @router.post("/login", name="login", status_code=200)
    def login(
        request: Request,
        email: Annotated[str, Body(example="info@openmined.org")],
        password: Annotated[str, Body(example="changethis")],
    ) -> Response:
        if TRACE_MODE:
            with trace.get_tracer(login.__module__).start_as_current_span(
                login.__qualname__,
                context=extract(request.headers),
                kind=trace.SpanKind.SERVER,
            ):
                return handle_login(email, password, worker)
        else:
            return handle_login(email, password, worker)

    @router.post("/register", name="register", status_code=200)
    def register(
        request: Request, data: Annotated[bytes, Depends(get_body)]
    ) -> Response:
        if TRACE_MODE:
            with trace.get_tracer(register.__module__).start_as_current_span(
                register.__qualname__,
                context=extract(request.headers),
                kind=trace.SpanKind.SERVER,
            ):
                return handle_register(data, worker)
        else:
            return handle_register(data, worker)

    return router



# future
from __future__ import annotations

# stdlib
import ast
from collections.abc import Callable
from copy import deepcopy
from enum import Enum
import hashlib
import inspect
from inspect import Parameter
from inspect import Signature
from io import StringIO
import sys
from typing import Any
from typing import ClassVar

# third party
from RestrictedPython import compile_restricted
from pydantic import field_validator
from pydantic import model_validator
import requests
from result import Err
from result import Ok
from result import Result

# relative
from ...abstract_server import ServerType
from ...client.api import APIRegistry
from ...client.api import RemoteFunction
from ...client.api import ServerIdentity
from ...serde.recursive_primitives import recursive_serde_register_type
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...store.document_store import PartitionKey
from ...types.datetime import DateTime
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import SyftObject
from ...types.syft_object_registry import SyftObjectRegistry
from ...types.transforms import TransformContext
from ...types.transforms import generate_id
from ...types.transforms import transform
from ...types.twin_object import TwinObject
from ...types.uid import UID
from ...util.util import is_interpreter_jupyter
from ..action.action_endpoint import CustomEndpointActionObject
from ..action.action_object import ActionObject
from ..action.action_permissions import ActionObjectPermission
from ..action.action_permissions import ActionPermission
from ..code.code_parse import GlobalsVisitor
from ..code.unparse import unparse
from ..context import AuthedServiceContext
from ..context import ChangeContext
from ..context import ServerServiceContext
from ..dataset.dataset import Asset
from ..response import SyftError
from ..response import SyftSuccess

DEFAULT_USER_POLICY_VERSION = 1

PolicyUserVerifyKeyPartitionKey = PartitionKey(
    key="user_verify_key", type_=SyftVerifyKey
)
PyCodeObject = Any


def extract_uid(v: Any) -> UID:
    value = v
    if isinstance(v, ActionObject):
        value = v.id
    if isinstance(v, TwinObject):
        value = v.id

    if not isinstance(value, UID):
        raise Exception(f"Input {v} must have a UID not {type(v)}")
    return value


def filter_only_uids(results: Any) -> list[UID] | dict[str, UID] | UID:
    # Prevent checking for __len__ on ActionObject (creates an Action)
    if isinstance(results, ActionObject):
        return extract_uid(results)

    if not hasattr(results, "__len__"):
        results = [results]

    if isinstance(results, list):
        output_list = [extract_uid(v) for v in results]
        return output_list
    elif isinstance(results, dict):
        output_dict = {}
        for k, v in results.items():
            output_dict[k] = extract_uid(v)
        return output_dict
    return extract_uid(results)


class Policy(SyftObject):
    # version
    __canonical_name__: str = "Policy"
    __version__ = SYFT_OBJECT_VERSION_1
    has_safe_serde: ClassVar[bool] = True

    id: UID
    init_kwargs: dict[Any, Any] = {}

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        if "init_kwargs" in kwargs:
            init_kwargs = kwargs["init_kwargs"]
            del kwargs["init_kwargs"]
        else:
            init_kwargs = deepcopy(kwargs)
            if "id" in init_kwargs:
                del init_kwargs["id"]
        super().__init__(init_kwargs=init_kwargs, *args, **kwargs)  # noqa: B026

    @property
    def policy_code(self) -> str:
        mro = reversed(type(self).mro())
        op_code = ""
        for klass in mro:
            if "Policy" in klass.__name__:
                op_code += inspect.getsource(klass)
                op_code += "\n"
        return op_code

    def is_valid(self, *args: list, **kwargs: dict) -> SyftSuccess | SyftError:  # type: ignore
        return SyftSuccess(message="Policy is valid.")

    def public_state(self) -> Any:
        raise NotImplementedError


@serializable(canonical_name="UserPolicyStatus", version=1)
class UserPolicyStatus(Enum):
    SUBMITTED = "submitted"
    DENIED = "denied"
    APPROVED = "approved"


def partition_by_server(kwargs: dict[str, Any]) -> dict[ServerIdentity, dict[str, UID]]:
    # relative
    from ...client.api import APIRegistry
    from ...client.api import RemoteFunction
    from ...client.api import ServerIdentity
    from ...types.twin_object import TwinObject
    from ..action.action_object import ActionObject

    # fetches the all the current api's connected
    api_list = APIRegistry.get_all_api()
    output_kwargs = {}
    for k, v in kwargs.items():
        uid = v
        if isinstance(v, ActionObject):
            uid = v.id
        if isinstance(v, TwinObject):
            uid = v.id
        if isinstance(v, RemoteFunction):
            uid = v.custom_function_actionobject_id()
        if isinstance(v, Asset):
            uid = v.action_id
        if not isinstance(uid, UID):
            raise Exception(f"Input {k} must have a UID not {type(v)}")

        _obj_exists = False
        for api in api_list:
            try:
                if api.services.action.exists(uid):
                    server_identity = ServerIdentity.from_api(api)
                    if server_identity not in output_kwargs:
                        output_kwargs[server_identity] = {k: uid}
                    else:
                        output_kwargs[server_identity].update({k: uid})

                    _obj_exists = True
                    break
            except requests.exceptions.ConnectionError:
                # To handle the cases , where there an old api objects in
                # in APIRegistry
                continue

        if not _obj_exists:
            raise Exception(f"Input data {k}:{uid} does not belong to any Datasite")

    return output_kwargs


@serializable()
class PolicyRule(SyftObject):
    __canonical_name__ = "PolicyRule"
    __version__ = SYFT_OBJECT_VERSION_1

    kw: str
    requires_input: bool = True

    def is_met(
        self, context: AuthedServiceContext, action_object: ActionObject
    ) -> bool:
        return False


@serializable()
class CreatePolicyRule(SyftObject):
    __canonical_name__ = "CreatePolicyRule"
    __version__ = SYFT_OBJECT_VERSION_1

    val: Any


@serializable()
class CreatePolicyRuleConstant(CreatePolicyRule):
    __canonical_name__ = "CreatePolicyRuleConstant"
    __version__ = SYFT_OBJECT_VERSION_1

    val: Any
    klass: None | type = None

    @model_validator(mode="before")
    @classmethod
    def set_klass(cls, data: Any) -> Any:
        val = data["val"]
        if isinstance(val, RemoteFunction):
            klass = CustomEndpointActionObject
        else:
            klass = type(val)
        data["klass"] = klass
        return data

    @field_validator("val", mode="after")
    @classmethod
    def idify_endpoints(cls, value: str) -> str:
        if isinstance(value, RemoteFunction):
            return value.custom_function_actionobject_id()
        return value

    def to_policy_rule(self, kw: Any) -> PolicyRule:
        return Constant(kw=kw, val=self.val, klass=self.klass)


@serializable()
class Matches(PolicyRule):
    __canonical_name__ = "Matches"
    __version__ = SYFT_OBJECT_VERSION_1

    val: UID

    def is_met(
        self, context: AuthedServiceContext, action_object: ActionObject
    ) -> bool:
        return action_object.id == self.val


@serializable()
class Constant(PolicyRule):
    __canonical_name__ = "PreFill"
    __version__ = SYFT_OBJECT_VERSION_1

    val: Any
    klass: type
    requires_input: bool = False

    def is_met(self, context: AuthedServiceContext, *args: Any, **kwargs: Any) -> bool:
        return True

    def transform_kwarg(
        self, context: AuthedServiceContext, val: Any
    ) -> Result[Any, str]:
        if isinstance(self.val, UID):
            if issubclass(self.klass, CustomEndpointActionObject):
                res = context.server.get_service("actionservice").get(context, self.val)
                if res.is_err():
                    return res
                else:
                    obj = res.ok()
                    return Ok(obj.syft_action_data)
        return Ok(self.val)


@serializable()
class UserOwned(PolicyRule):
    __canonical_name__ = "UserOwned"
    __version__ = SYFT_OBJECT_VERSION_1

    # str, float, int, bool, dict, list, set, tuple

    type: (
        type[str]
        | type[float]
        | type[int]
        | type[bool]
        | type[dict]
        | type[list]
        | type[set]
        | type[tuple]
        | None
    )

    def is_owned(
        self, context: AuthedServiceContext, action_object: ActionObject
    ) -> bool:
        action_store = context.server.get_service("actionservice").store
        return action_store.has_permission(
            ActionObjectPermission(
                action_object.id, ActionPermission.OWNER, context.credentials
            )
        )

    def is_met(
        self, context: AuthedServiceContext, action_object: ActionObject
    ) -> bool:
        return type(action_object.syft_action_data) == self.type and self.is_owned(
            context, action_object
        )


def user_code_arg2id(arg: Any) -> UID:
    if isinstance(arg, ActionObject):
        uid = arg.id
    elif isinstance(arg, TwinObject):
        uid = arg.id
    elif isinstance(arg, Asset):
        uid = arg.action_id
    elif isinstance(arg, RemoteFunction):
        # TODO: Beach Fix
        # why do we need another call to the server to get the UID?
        uid = arg.custom_function_actionobject_id()
    else:
        uid = arg
    return uid


def retrieve_item_from_db(id: UID, context: AuthedServiceContext) -> ActionObject:
    # relative
    from ...service.action.action_object import TwinMode

    action_service = context.server.get_service("actionservice")
    root_context = AuthedServiceContext(
        server=context.server, credentials=context.server.verify_key
    )
    value = action_service._get(
        context=root_context,
        uid=id,
        twin_mode=TwinMode.NONE,
        has_permission=True,
    )
    if value.is_err():
        return value
    else:
        return value.ok()


class InputPolicy(Policy):
    __canonical_name__ = "InputPolicy"
    __version__ = SYFT_OBJECT_VERSION_1

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        if "init_kwargs" in kwargs:
            init_kwargs = kwargs["init_kwargs"]
            del kwargs["init_kwargs"]
        else:
            # TODO: remove this tech debt, dont remove the id mapping functionality
            init_kwargs = partition_by_server(kwargs)
        super().__init__(*args, init_kwargs=init_kwargs, **kwargs)

    def _is_valid(
        self,
        context: AuthedServiceContext,
        usr_input_kwargs: dict,
        code_item_id: UID,
    ) -> Result[bool, str]:
        raise NotImplementedError

    def filter_kwargs(
        self,
        kwargs: dict[Any, Any],
        context: AuthedServiceContext,
        code_item_id: UID,
    ) -> dict[Any, Any]:
        raise NotImplementedError

    @property
    def inputs(self) -> dict[ServerIdentity, Any]:
        return self.init_kwargs

    def _inputs_for_context(self, context: ChangeContext) -> dict | SyftError:
        user_server_view = ServerIdentity.from_change_context(context)
        inputs = self.inputs[user_server_view]
        root_context = AuthedServiceContext(
            server=context.server, credentials=context.approving_user_credentials
        ).as_root_context()

        action_service = context.server.get_service("actionservice")
        for var_name, uid in inputs.items():
            action_object = action_service.get(uid=uid, context=root_context)
            if action_object.is_err():
                return SyftError(message=action_object.err())
            action_object_value = action_object.ok()
            # resolve syft action data from blob store
            if isinstance(action_object_value, TwinObject):
                action_object_value.private_obj.syft_action_data  # noqa: B018
                action_object_value.mock_obj.syft_action_data  # noqa: B018
            elif isinstance(action_object_value, ActionObject):
                action_object_value.syft_action_data  # noqa: B018
            inputs[var_name] = action_object_value
        return inputs


@serializable()
class MixedInputPolicy(InputPolicy):
    # version
    __canonical_name__ = "MixedInputPolicy"
    __version__ = SYFT_OBJECT_VERSION_1

    kwarg_rules: dict[ServerIdentity, dict[str, PolicyRule]]

    def __init__(
        self, init_kwargs: Any = None, client: Any = None, *args: Any, **kwargs: Any
    ) -> None:
        if init_kwargs is not None:
            kwarg_rules = init_kwargs
            kwargs = {}
        else:
            server_identity = self.find_server_identity(kwargs, client)
            kwarg_rules_current_server = {}
            for kw, arg in kwargs.items():
                if isinstance(
                    arg, UID | Asset | ActionObject | TwinObject | RemoteFunction
                ):
                    kwarg_rules_current_server[kw] = Matches(
                        kw=kw, val=user_code_arg2id(arg)
                    )
                elif arg in [str, float, int, bool, dict, list, set, tuple]:  # type: ignore[unreachable]
                    kwarg_rules_current_server[kw] = UserOwned(kw=kw, type=arg)
                elif isinstance(arg, CreatePolicyRule):
                    kwarg_rules_current_server[kw] = arg.to_policy_rule(kw)
                else:
                    raise ValueError("Incorrect argument")
            kwarg_rules = {server_identity: kwarg_rules_current_server}

        super().__init__(
            *args, kwarg_rules=kwarg_rules, init_kwargs=kwarg_rules, **kwargs
        )

    def transform_kwargs(
        self, context: AuthedServiceContext, kwargs: dict[str, Any]
    ) -> dict[str, Any]:
        for _, rules in self.kwarg_rules.items():
            for kw, rule in rules.items():
                if hasattr(rule, "transform_kwarg"):
                    res_val = rule.transform_kwarg(context, kwargs.get(kw, None))
                    if res_val.is_err():
                        return res_val
                    else:
                        kwargs[kw] = res_val.ok()
        return Ok(kwargs)

    def find_server_identity(
        self, kwargs: dict[str, Any], client: Any = None
    ) -> ServerIdentity:
        if client is not None:
            return ServerIdentity.from_api(client.api)

        apis = APIRegistry.get_all_api()
        matches = set()
        has_ids = False
        for val in kwargs.values():
            # we mostly get the UID here because we don't want to store all those
            # other objects, so we need to create a global UID obj lookup service
            if isinstance(
                val, UID | Asset | ActionObject | TwinObject | RemoteFunction
            ):
                has_ids = True
                id = user_code_arg2id(val)
                for api in apis:
                    # TODO: Beach Fix
                    # here be dragons, we need to refactor this since the existance
                    # depends on the type and service
                    # also the whole ServerIdentity needs to be removed
                    check_endpoints = [
                        api.services.action.exists,
                        api.services.api.exists,
                    ]
                    for check_endpoint in check_endpoints:
                        result = check_endpoint(id)
                        if result:
                            break  # stop looking
                    if result:
                        server_identity = ServerIdentity.from_api(api)
                        matches.add(server_identity)

        if len(matches) == 0:
            if not has_ids:
                if len(apis) == 1:
                    return ServerIdentity.from_api(api)
                else:
                    raise ValueError(
                        "Multiple Server Identities, please only login to one client (for this policy) and try again"
                    )
            else:
                raise ValueError("No Server Identities")
        if len(matches) > 1:
            # TODO: Beach Fix
            raise ValueError("Multiple Server Identities")
            # we need to fix this as its possible we could
            # grab the wrong API and call a different user context in jupyter testing
            pass  # just grab the first one
        return matches.pop()

    def filter_kwargs(
        self,
        kwargs: dict[str, UID],
        context: AuthedServiceContext,
        code_item_id: UID,
    ) -> Result[dict[Any, Any], str]:
        try:
            res = {}
            for _, rules in self.kwarg_rules.items():
                for kw, rule in rules.items():
                    if rule.requires_input:
                        passed_id = kwargs[kw]
                        actionobject: ActionObject = retrieve_item_from_db(
                            passed_id, context
                        )
                        rule_check_args = (actionobject,)
                    else:
                        rule_check_args = ()  # type: ignore
                        # TODO
                        actionobject = rule.value
                    if not rule.is_met(context, *rule_check_args):
                        raise ValueError(f"{rule} is not met")
                    else:
                        res[kw] = actionobject
        except Exception as e:
            return Err(str(e))
        return Ok(res)

    def _is_valid(
        self,
        context: AuthedServiceContext,
        usr_input_kwargs: dict,
        code_item_id: UID,
    ) -> Result[bool, str]:
        filtered_input_kwargs = self.filter_kwargs(
            kwargs=usr_input_kwargs,
            context=context,
            code_item_id=code_item_id,
        )

        if filtered_input_kwargs.is_err():
            return filtered_input_kwargs

        filtered_input_kwargs = filtered_input_kwargs.ok()

        expected_input_kwargs = set()
        for _inp_kwargs in self.inputs.values():
            for k in _inp_kwargs.keys():
                if k not in usr_input_kwargs:
                    return Err(f"Function missing required keyword argument: '{k}'")
            expected_input_kwargs.update(_inp_kwargs.keys())

        permitted_input_kwargs = list(filtered_input_kwargs.keys())
        not_approved_kwargs = set(expected_input_kwargs) - set(permitted_input_kwargs)
        if len(not_approved_kwargs) > 0:
            return Err(
                f"Input arguments: {not_approved_kwargs} to the function are not approved yet."
            )
        return Ok(True)


def retrieve_from_db(
    code_item_id: UID, allowed_inputs: dict[str, UID], context: AuthedServiceContext
) -> Result[dict[str, Any], str]:
    # relative
    from ...service.action.action_object import TwinMode

    action_service = context.server.get_service("actionservice")
    code_inputs = {}

    # When we are retrieving the code from the database, we need to use the server's
    # verify key as the credentials. This is because when we approve the code, we
    # we allow the private data to be used only for this specific code.
    # but we are not modifying the permissions of the private data

    root_context = AuthedServiceContext(
        server=context.server, credentials=context.server.verify_key
    )
    if context.server.server_type == ServerType.DATASITE:
        for var_name, arg_id in allowed_inputs.items():
            kwarg_value = action_service._get(
                context=root_context,
                uid=arg_id,
                twin_mode=TwinMode.NONE,
                has_permission=True,
            )
            if kwarg_value.is_err():
                return Err(kwarg_value.err())
            code_inputs[var_name] = kwarg_value.ok()
    else:
        raise Exception(
            f"Invalid Server Type for Code Submission:{context.server.server_type}"
        )
    return Ok(code_inputs)


def allowed_ids_only(
    allowed_inputs: dict[ServerIdentity, Any],
    kwargs: dict[str, Any],
    context: AuthedServiceContext,
) -> dict[ServerIdentity, UID]:
    if context.server.server_type == ServerType.DATASITE:
        server_identity = ServerIdentity(
            server_name=context.server.name,
            server_id=context.server.id,
            verify_key=context.server.signing_key.verify_key,
        )
        allowed_inputs = allowed_inputs.get(server_identity, {})
    else:
        raise Exception(
            f"Invalid Server Type for Code Submission:{context.server.server_type}"
        )
    filtered_kwargs = {}
    for key in allowed_inputs.keys():
        if key in kwargs:
            value = kwargs[key]
            uid = value
            if not isinstance(uid, UID):
                uid = getattr(value, "id", None)

            if uid != allowed_inputs[key]:
                raise Exception(
                    f"Input with uid: {uid} for `{key}` not in allowed inputs: {allowed_inputs}"
                )
            filtered_kwargs[key] = value
    return filtered_kwargs


@serializable()
class ExactMatch(InputPolicy):
    # version
    __canonical_name__ = "ExactMatch"
    __version__ = SYFT_OBJECT_VERSION_1

    def filter_kwargs(
        self,
        kwargs: dict[Any, Any],
        context: AuthedServiceContext,
        code_item_id: UID,
    ) -> Result[dict[Any, Any], str]:
        try:
            allowed_inputs = allowed_ids_only(
                allowed_inputs=self.inputs, kwargs=kwargs, context=context
            )

            results = retrieve_from_db(
                code_item_id=code_item_id,
                allowed_inputs=allowed_inputs,
                context=context,
            )
        except Exception as e:
            return Err(str(e))
        return results

    def _is_valid(
        self,
        context: AuthedServiceContext,
        usr_input_kwargs: dict,
        code_item_id: UID,
    ) -> Result[bool, str]:
        filtered_input_kwargs = self.filter_kwargs(
            kwargs=usr_input_kwargs,
            context=context,
            code_item_id=code_item_id,
        )

        if filtered_input_kwargs.is_err():
            return filtered_input_kwargs

        filtered_input_kwargs = filtered_input_kwargs.ok()

        expected_input_kwargs = set()
        for _inp_kwargs in self.inputs.values():
            for k in _inp_kwargs.keys():
                if k not in usr_input_kwargs:
                    return Err(f"Function missing required keyword argument: '{k}'")
            expected_input_kwargs.update(_inp_kwargs.keys())

        permitted_input_kwargs = list(filtered_input_kwargs.keys())
        not_approved_kwargs = set(expected_input_kwargs) - set(permitted_input_kwargs)
        if len(not_approved_kwargs) > 0:
            return Err(
                f"Function arguments: {not_approved_kwargs} are not approved yet."
            )
        return Ok(True)


@serializable()
class OutputHistory(SyftObject):
    # version
    __canonical_name__ = "OutputHistory"
    __version__ = SYFT_OBJECT_VERSION_1

    output_time: DateTime
    outputs: list[UID] | dict[str, UID] | None = None
    executing_user_verify_key: SyftVerifyKey


class OutputPolicy(Policy):
    # version
    __canonical_name__ = "OutputPolicy"
    __version__ = SYFT_OBJECT_VERSION_1

    output_kwargs: list[str] = []
    server_uid: UID | None = None
    output_readers: list[SyftVerifyKey] = []

    def apply_to_output(
        self,
        context: ServerServiceContext,
        outputs: Any,
        update_policy: bool = True,
    ) -> Any:
        # output_uids: Union[Dict[str, Any], list] = filter_only_uids(outputs)
        # if isinstance(output_uids, UID):
        #     output_uids = [output_uids]
        # history = OutputHistory(
        #     output_time=DateTime.now(),
        #     outputs=output_uids,
        #     executing_user_verify_key=context.credentials,
        # )
        # self.output_history.append(history)

        return outputs

    def is_valid(self, context: AuthedServiceContext) -> SyftSuccess | SyftError:  # type: ignore
        raise NotImplementedError()


@serializable()
class OutputPolicyExecuteCount(OutputPolicy):
    __canonical_name__ = "OutputPolicyExecuteCount"
    __version__ = SYFT_OBJECT_VERSION_1

    limit: int

    @property
    def count(self) -> SyftError | int:
        api = APIRegistry.api_for(
            self.syft_server_location, self.syft_client_verify_key
        )
        if api is None:
            raise ValueError(
                f"api is None. You must login to {self.syft_server_location}"
            )
        output_history = api.services.output.get_by_output_policy_id(self.id)

        if isinstance(output_history, SyftError):
            return output_history
        return len(output_history)

    @property
    def is_valid(self) -> SyftSuccess | SyftError:  # type: ignore
        execution_count = self.count
        is_valid = execution_count < self.limit
        if is_valid:
            return SyftSuccess(
                message=f"Policy is still valid. count: {execution_count} < limit: {self.limit}"
            )
        return SyftError(
            message=f"Policy is no longer valid. count: {execution_count} >= limit: {self.limit}"
        )

    def _is_valid(self, context: AuthedServiceContext) -> SyftSuccess | SyftError:
        output_service = context.server.get_service("outputservice")
        output_history = output_service.get_by_output_policy_id(context, self.id)
        if isinstance(output_history, SyftError):
            return output_history
        execution_count = len(output_history)

        is_valid = execution_count < self.limit
        if is_valid:
            return SyftSuccess(
                message=f"Policy is still valid. count: {execution_count} < limit: {self.limit}"
            )
        return SyftError(
            message=f"Policy is no longer valid. count: {execution_count} >= limit: {self.limit}"
        )

    def public_state(self) -> dict[str, int]:
        return {"limit": self.limit, "count": self.count}


@serializable()
class OutputPolicyExecuteOnce(OutputPolicyExecuteCount):
    __canonical_name__ = "OutputPolicyExecuteOnce"
    __version__ = SYFT_OBJECT_VERSION_1

    limit: int = 1


SingleExecutionExactOutput = OutputPolicyExecuteOnce


@serializable(canonical_name="CustomPolicy", version=1)
class CustomPolicy(type):
    # capture the init_kwargs transparently
    def __call__(cls, *args: Any, **kwargs: Any) -> None:
        obj = super().__call__(*args, **kwargs)
        obj.init_kwargs = kwargs
        return obj


recursive_serde_register_type(CustomPolicy, canonical_name="CustomPolicy", version=1)


@serializable(canonical_name="CustomOutputPolicy", version=1)
class CustomOutputPolicy(metaclass=CustomPolicy):
    def apply_to_output(
        self,
        context: ServerServiceContext,
        outputs: Any,
        update_policy: bool = True,
    ) -> Any | None:
        return outputs


class UserOutputPolicy(OutputPolicy):
    __canonical_name__ = "UserOutputPolicy"

    # Do not validate private attributes of user-defined policies, User annotations can
    # contain any type and throw a NameError when resolving.
    __validate_private_attrs__ = False
    pass


class UserInputPolicy(InputPolicy):
    __canonical_name__ = "UserInputPolicy"
    __validate_private_attrs__ = False
    pass


@serializable()
class EmpyInputPolicy(InputPolicy):
    __canonical_name__ = "EmptyInputPolicy"
    pass


class CustomInputPolicy(metaclass=CustomPolicy):
    pass


@serializable()
class UserPolicy(Policy):
    __canonical_name__: str = "UserPolicy"
    __version__ = SYFT_OBJECT_VERSION_1
    has_safe_serde: ClassVar[bool] = False

    id: UID
    server_uid: UID | None = None
    user_verify_key: SyftVerifyKey
    raw_code: str
    parsed_code: str
    signature: inspect.Signature
    class_name: str
    unique_name: str
    code_hash: str
    status: UserPolicyStatus = UserPolicyStatus.SUBMITTED

    # TODO: fix the mypy issue
    @property  # type: ignore
    def byte_code(self) -> PyCodeObject | None:
        return compile_byte_code(self.parsed_code)

    @property
    def policy_code(self) -> str:
        return self.raw_code

    def apply_to_output(
        self,
        context: ServerServiceContext,
        outputs: Any,
        update_policy: bool = True,
    ) -> Any | None:
        return outputs


def new_getfile(object: Any) -> Any:  # TODO: fix the mypy issue
    if not inspect.isclass(object):
        return inspect.getfile(object)

    # Lookup by parent module (as in current inspect)
    if hasattr(object, "__module__"):
        object_ = sys.modules.get(object.__module__)
        if object_ is not None and hasattr(object_, "__file__"):
            return object_.__file__

    # If parent module is __main__, lookup by methods (NEW)
    for _, member in inspect.getmembers(object):
        if (
            inspect.isfunction(member)
            and object.__qualname__ + "." + member.__name__ == member.__qualname__
        ):
            return inspect.getfile(member)
    else:
        raise TypeError(f"Source for {object!r} not found")


def get_code_from_class(policy: type[CustomPolicy]) -> str:
    klasses = [inspect.getmro(policy)[0]]  #
    whole_str = ""
    for klass in klasses:
        if is_interpreter_jupyter():
            # third party
            from IPython.core.magics.code import extract_symbols

            cell_code = "".join(inspect.linecache.getlines(new_getfile(klass)))
            class_code = extract_symbols(cell_code, klass.__name__)[0][0]
        else:
            class_code = inspect.getsource(klass)
        whole_str += class_code
    return whole_str


@serializable()
class SubmitUserPolicy(Policy):
    __canonical_name__ = "SubmitUserPolicy"
    __version__ = SYFT_OBJECT_VERSION_1

    id: UID | None = None  # type: ignore[assignment]
    code: str
    class_name: str
    input_kwargs: list[str]

    def compile(self) -> PyCodeObject:
        return compile_restricted(self.code, "<string>", "exec")

    @staticmethod
    def from_obj(policy_obj: CustomPolicy) -> SubmitUserPolicy:
        user_class = policy_obj.__class__
        init_f_code = user_class.__init__.__code__
        return SubmitUserPolicy(
            code=get_code_from_class(user_class),
            class_name=user_class.__name__,
            input_kwargs=init_f_code.co_varnames[1 : init_f_code.co_argcount],
        )


def hash_code(context: TransformContext) -> TransformContext:
    if context.output is None:
        return context
    code = context.output["code"]
    del context.output["code"]
    context.output["raw_code"] = code
    code_hash = hashlib.sha256(code.encode("utf8")).hexdigest()
    context.output["code_hash"] = code_hash

    return context


def generate_unique_class_name(context: TransformContext) -> TransformContext:
    # TODO: Do we need to check if the initial name contains underscores?
    if context.output is not None:
        code_hash = context.output["code_hash"]
        service_class_name = context.output["class_name"]
        unique_name = f"{service_class_name}_{context.credentials}_{code_hash}"
        context.output["unique_name"] = unique_name
    else:
        raise ValueError(f"{context}'s output is None. No transformation happened")

    return context


def compile_byte_code(parsed_code: str) -> PyCodeObject | None:
    try:
        return compile(parsed_code, "<string>", "exec")
    except Exception as e:
        print("WARNING: to compile byte code", e)
    return None


def process_class_code(raw_code: str, class_name: str) -> str:
    tree = ast.parse(raw_code)
    v = GlobalsVisitor()
    v.visit(tree)
    if len(tree.body) != 1 or not isinstance(tree.body[0], ast.ClassDef):
        raise Exception(
            "Class code should only contain the Class definition for your policy"
        )
    old_class = tree.body[0]
    if len(old_class.bases) != 1 or old_class.bases[0].attr not in [
        CustomInputPolicy.__name__,
        CustomOutputPolicy.__name__,
    ]:
        raise Exception(
            f"Class code should either implement {CustomInputPolicy.__name__} "
            f"or {CustomOutputPolicy.__name__}"
        )

    # TODO: changes the bases
    old_class.bases[0].attr = old_class.bases[0].attr.replace("Custom", "User")

    serializable_name = ast.Name(id="sy.serializable", ctx=ast.Load())
    serializable_decorator = ast.Call(
        func=serializable_name,
        args=[],
        keywords=[],
    )

    new_class = tree.body[0]
    # TODO add this manually
    for stmt in new_class.body:
        if isinstance(stmt, ast.FunctionDef) and stmt.name == "__init__":
            stmt.name = "__user_init__"

    # change the module that the code will reference
    # this is required for the @serializable to mount it in the right path for serde
    new_line = ast.parse('__module__ = "syft.user"')
    new_class.body.append(new_line.body[0])
    new_line = ast.parse(f'__canonical_name__ = "{class_name}"')
    new_class.body.append(new_line.body[0])
    new_line = ast.parse("__version__ = 1")
    new_class.body.append(new_line.body[0])
    new_class.name = class_name
    new_class.decorator_list = [serializable_decorator]
    new_body = []
    new_body.append(
        ast.ImportFrom(
            module="__future__",
            names=[ast.alias(name="annotations", asname="annotations")],
            level=0,
        )
    )
    new_body.append(ast.Import(names=[ast.alias(name="syft", asname="sy")], level=0))
    typing_types = [
        "Any",
        "Callable",
        "ClassVar",
        "Dict",
        "List",
        "Optional",
        "Set",
        "Tuple",
        "Type",
    ]
    new_body.append(
        ast.ImportFrom(
            module="typing",
            names=[
                ast.alias(name=typing_type, asname=typing_type)
                for typing_type in typing_types
            ],
            level=0,
        )
    )
    new_body.append(new_class)
    module = ast.Module(new_body, type_ignores=[])
    try:
        return unparse(module)
    except Exception as e:
        print("failed to unparse", e)
        raise e


def check_class_code(context: TransformContext) -> TransformContext:
    # TODO: define the proper checking for this case based on the ideas from UserCode
    # check for no globals
    # check for Policy template -> __init__, apply_to_output, public_state
    # parse init signature
    # check dangerous libraries, maybe compile_restricted already does that
    if context.output is None:
        return context

    try:
        processed_code = process_class_code(
            raw_code=context.output["raw_code"],
            class_name=context.output["unique_name"],
        )
        context.output["parsed_code"] = processed_code
    except Exception as e:
        raise e

    return context


def compile_code(context: TransformContext) -> TransformContext:
    if context.output is not None:
        byte_code = compile_byte_code(context.output["parsed_code"])
        if byte_code is None:
            raise Exception(
                "Unable to compile byte code from parsed code. "
                + context.output["parsed_code"]
            )
    else:
        raise ValueError(f"{context}'s output is None. No transformation happened")

    return context


def add_credentials_for_key(key: str) -> Callable:
    def add_credentials(context: TransformContext) -> TransformContext:
        if context.output is not None:
            context.output[key] = context.credentials

        return context

    return add_credentials


def generate_signature(context: TransformContext) -> TransformContext:
    if context.output is None:
        return context

    params = [
        Parameter(name=k, kind=Parameter.POSITIONAL_OR_KEYWORD)
        for k in context.output["input_kwargs"]
    ]
    sig = Signature(parameters=params)
    context.output["signature"] = sig

    return context


@transform(SubmitUserPolicy, UserPolicy)
def submit_policy_code_to_user_code() -> list[Callable]:
    return [
        generate_id,
        hash_code,
        generate_unique_class_name,
        generate_signature,
        check_class_code,
        # compile_code, # don't compile until approved
        add_credentials_for_key("user_verify_key"),
    ]


def register_policy_class(klass: type, unique_name: str) -> None:
    nonrecursive = False
    _serialize = None
    _deserialize = None
    attributes = list(klass.model_fields.keys())
    exclude_attrs: list = []
    serde_overrides: dict = {}
    hash_exclude_attrs: list = []
    cls = klass
    attribute_types: list = []
    version = 1

    serde_attributes = (
        nonrecursive,
        _serialize,
        _deserialize,
        attributes,
        exclude_attrs,
        serde_overrides,
        hash_exclude_attrs,
        cls,
        attribute_types,
        version,
    )

    SyftObjectRegistry.register_cls(
        canonical_name=unique_name, version=version, serde_attributes=serde_attributes
    )


def execute_policy_code(user_policy: UserPolicy) -> Any:
    stdout_ = sys.stdout
    stderr_ = sys.stderr

    try:
        stdout = StringIO()
        stderr = StringIO()

        sys.stdout = stdout
        sys.stderr = stderr

        class_name = user_policy.unique_name

        try:
            policy_class = SyftObjectRegistry.get_serde_class(
                class_name, version=DEFAULT_USER_POLICY_VERSION
            )
        except Exception:
            exec(user_policy.byte_code)  # nosec
            policy_class = eval(user_policy.unique_name)  # nosec

        register_policy_class(policy_class, user_policy.unique_name)

        sys.stdout = stdout_
        sys.stderr = stderr_

        return policy_class

    except Exception as e:
        print("execute_byte_code failed", e, file=stderr_)

    finally:
        sys.stdout = stdout_
        sys.stderr = stderr_


def load_policy_code(user_policy: UserPolicy) -> Any:
    try:
        policy_class = execute_policy_code(user_policy)
        return policy_class
    except Exception as e:
        raise Exception(f"Exception loading code. {user_policy}. {e}")


def init_policy(user_policy: UserPolicy, init_args: dict[str, Any]) -> Any:
    policy_class = load_policy_code(user_policy)
    policy_object = policy_class()

    # Unwrapp {ServerIdentity : {x: y}} -> {x: y}
    # Tech debt : For input policies, we required to have ServerIdentity args beforehand,
    # therefore at this stage we had to return back to the normal args.
    # Maybe there's better way to do it.
    if len(init_args) and isinstance(list(init_args.keys())[0], ServerIdentity):
        unwrapped_init_kwargs = init_args
        if len(init_args) > 1:
            raise Exception("You shoudn't have more than one Server Identity.")
        # Otherwise, unwrapp it
        init_args = init_args[list(init_args.keys())[0]]

    init_args = {k: v for k, v in init_args.items() if k != "id"}

    # For input policies, this initializer wouldn't work properly:
    # 1 - Passing {ServerIdentity: {kwargs:UIDs}} as keyword args doesn't work since keys must be strings
    # 2 - Passing {kwargs: UIDs} in this initializer would not trigger the partition servers from the
    # InputPolicy initializer.
    # The cleanest way to solve it is by checking if it's an Input Policy, and then, setting it manually.
    policy_object.__user_init__(**init_args)
    if isinstance(policy_object, InputPolicy):
        policy_object.init_kwargs = unwrapped_init_kwargs
    return policy_object
